"""Various metrics used to evaluate clustering."""

import numba as nb
import numpy as np

from spine.math.linalg import contingency_table
from spine.math.metrics import adjusted_mutual_info_score, adjusted_rand_score

__all__ = ["pur", "eff", "pur_eff", "ari", "ami", "sbd"]


def pur(truth, pred, batch_ids=None, per_cluster=True):
    """Assignment purity.

    Parameters
    ----------
    truth : np.ndarray
        (N) Set of true labels
    pred : np.ndarray
        (N) Set of predicted labels
    batch_ids : np.ndarray, optional
        (N) Batch IDs
    per_cluster : bool, default True
        If `True`, computes the purity per predicted cluster, than averages it

    Returns
    -------
    float
        Assignment purity
    """
    # If the vectors compared are empty, nothing to do
    if len(truth) == 0:
        return -1.0

    # Transform labels to be unique across all batch entries
    truth, _, truth_counts = unique_labels(truth, batch_ids)
    pred, _, pred_counts = unique_labels(pred, batch_ids)

    # Compute the contingency table
    table = contingency_table(truth, pred, len(truth_counts), len(pred_counts))

    # Evaluate the purity for each predicted cluster
    if per_cluster:
        purities = table.max(axis=0) / pred_counts
        return purities.mean()

    else:
        purity = np.sum(table.max(axis=0)) / len(pred)
        return purity


def eff(truth, pred, batch_ids=None, per_cluster=True):
    """Assignment efficiency, evaluated per true cluster and averaged.

    Parameters
    ----------
    truth : np.ndarray
        (N) Set of true labels
    pred : np.ndarray
        (N) Set of predicted labels
    batch_ids : np.ndarray, optional
        (N) Batch IDs
    per_cluster : bool, default True
        If `True`, computes the efficiency per truth cluster, than averages it

    Returns
    -------
    float
        Assignment efficiency
    """
    # If the vectors compared are empty, nothing to do
    if len(truth) == 0:
        return -1.0

    # Transform labels to be unique across all batch entries
    truth, _, truth_counts = unique_labels(truth, batch_ids)
    pred, _, pred_counts = unique_labels(pred, batch_ids)

    # Compute the contingency table
    table = contingency_table(truth, pred, len(truth_counts), len(pred_counts))

    # Evaluate the efficiency for each true cluster
    if per_cluster:
        efficiencies = table.max(axis=1) / truth_counts
        return efficiencies.mean()

    else:
        efficiency = np.sum(table.max(axis=1)) / len(truth)
        return efficiency


def pur_eff(truth, pred, batch_ids=None, per_cluster=True):
    """Assignment purity and efficiency.

    Parameters
    ----------
    truth : np.ndarray
        (N) Set of true labels
    pred : np.ndarray
        (N) Set of predicted labels
    batch_ids : np.ndarray, optional
        (N) Batch IDs
    per_cluster : bool, default True
        If `True`, computes the metrics per predicted cluster, than averages them

    Returns
    -------
    float
        Assignment purity
    float
        Assignment efficiency
    """
    # If the vectors compared are empty, nothing to do
    if len(truth) == 0:
        return -1.0, -1.0

    # Transform labels to be unique across all batch entries
    truth, _, truth_counts = unique_labels(truth, batch_ids)
    pred, _, pred_counts = unique_labels(pred, batch_ids)

    # Compute the contingency table
    table = contingency_table(truth, pred, len(truth_counts), len(pred_counts))

    # Evaluate the purity and efficiency
    if per_cluster:
        purities = table.max(axis=0) / pred_counts
        efficiencies = table.max(axis=1) / truth_counts

        return purities.mean(), efficiencies.mean()

    else:
        purity = np.sum(table.max(axis=0)) / len(pred)
        efficiency = np.sum(table.max(axis=1)) / len(truth)

        return purity, efficiency


def ari(truth, pred, batch_ids=None):
    """Computes the Adjusted Rand Index (ARI) between two sets of labels.

    Parameters
    ----------
    truth : np.ndarray
        (N) Set of true labels
    pred : np.ndarray
        (N) Set of predicted labels
    batch_ids : np.ndarray, optional
        (N) Batch IDs

    Returns
    -------
    float
        Adjusted Rand Index (ARI) value
    """
    # If the vectors compared are empty, nothing to do
    if len(truth) == 0:
        return -1.0

    # If required, transform labels to be unique across all batch entries
    if batch_ids is not None:
        truth = unique_labels(truth, batch_ids)[0]
        pred = unique_labels(pred, batch_ids)[0]

    return adjusted_rand_score(truth, pred)


def ami(truth, pred, batch_ids=None):
    """Computes the Adjusted Mutual Information (AMI) between two sets of labels.

    Parameters
    ----------
    truth : np.ndarray
        (N) Set of true labels
    pred : np.ndarray
        (N) Set of predicted labels
    batch_ids : np.ndarray, optional
        (N) Batch IDs

    Returns
    -------
    float
        Adjusted Mutual Information (AMI) value
    """
    # If the vectors compared are empty, nothing to do
    if len(truth) == 0:
        return -1.0

    # If required, transform labels to be unique across all batch entries
    if batch_ids is not None:
        truth = unique_labels(truth, batch_ids)[0]
        pred = unique_labels(pred, batch_ids)[0]

    return adjusted_mutual_info_score(truth, pred)


def sbd(truth, pred, batch_ids=None):
    """Compute the Symmetric Best Dice (SBD) score between two sets of labels.

    Parameters
    ----------
    truth : np.ndarray
        (N) Set of true labels
    pred : np.ndarray
        (N) Set of predicted labels
    batch_ids : np.ndarray, optional
        (N) Batch IDs

    Returns
    -------
    float
        Symmetric best dice value
    """
    # Transform labels to be unique across all batch entries
    truth, truth_unique, truth_counts = unique_labels(truth, batch_ids)
    pred, pred_unique, pred_counts = unique_labels(pred, batch_ids)

    # Compute the best dice both ways, take the minimum as the symmetric score
    bd1 = bd(truth, truth_unique, truth_counts, pred, pred_unique, pred_counts)
    bd2 = bd(pred, pred_unique, pred_counts, truth, truth_unique, truth_counts)

    return min(bd1, bd2)


def bd(truth, truth_unique, truth_counts, pred, pred_unique, pred_counts):
    """Computes the Best Dice (BD) between two sets of labels.

    Parameters
    ----------
    truth : np.ndarray
        (N) Set of true labels
    truth_unique : np.ndarray
        (K) Set of unique true labels
    truth_counts : np.ndarray
        (K) Number of realization of each unique true label
    pred : np.ndarray
        (N) Set of predicted labels
    pred_unique : np.ndarray
        (L) Set of unique predicted labels
    pred_counts : np.ndarray
        (L) Number of realization of each unique predicted label
    """
    # If the vectors compared are empty, nothing to do
    if len(truth) == 0:
        return -1.0

    # Loop over the predicted clusters
    total_bd = 0.0
    for i, c in enumerate(pred_unique):
        # Get the composition of the predicted cluster in the label array
        unique, counts = np.unique(truth[pred == c], return_counts=True)

        # Compute the best dice for this cluster
        best_dice = 0.0
        for j, d in enumerate(unique):
            dice = 2 * counts[j] / (pred_counts[i] + truth_counts[d])
            if dice > best_dice:
                best_dice = dice

        # Increment
        total_bd += best_dice

    # Take the mean best dice as a clustering score
    return total_bd / len(pred_unique)


def unique_labels(labels, batch_ids=None):
    """Transforms labels to range from 0 to C-1 labels (with C the number of
    unique values in the label array.

    If batch IDs are provided, ensures that the labels are unique at the batch
    level as well.

    Parameters
    ----------
    labels : np.ndarray
        (N) Labels
    batch_ids : np.ndarray, optional
        (N) Batch IDs

    Returns
    -------
    inverse : np.ndarray
        (N) Unique labels across all entries in the batch
    unique : np.ndarray
        (C) Unique set of labels
    counts : np.ndarray
        (C) Number of labels which belong to each unique category
    """
    if batch_ids is not None:
        labels = np.stack((labels, batch_ids))
    unique, inverse, counts = np.unique(
        labels, axis=-1, return_inverse=True, return_counts=True
    )

    return inverse, unique, counts



"""Module with functions that operate on collections of pixels (clusters).

A cluster is typically represented as a list of row indexes pointing at the
voxels that up the cluster out of a tensor of pixels.
"""

import numba as nb
import numpy as np

import spine.math as sm
from spine.data import IndexBatch, TensorBatch
from spine.utils.conditional import torch
from spine.utils.globals import (
    CLUST_COL,
    COORD_COLS,
    COORD_END_COLS,
    COORD_START_COLS,
    COORD_TIME_COL,
    GROUP_COL,
    PART_COL,
    SHAPE_COL,
    VALUE_COL,
)
from spine.utils.jit import numbafy


def form_clusters_batch(data, min_size=-1, column=CLUST_COL, shapes=None):
    """Batched version of :func:`form_clusters`.

    Parameters
    ----------
    data : TensorBatch
        Batch of cluster label data tensor
    min_size : int, default -1
        Minimum size of a cluster to be included in the list
    column : int, default CLUST_COL
        Column of the label tensor to use to fetch the pixel cluster IDs
    shapes : List[int], optional
        List of semantic classes to include in the list of cluster

    Returns
    -------
    IndexBatch
        Object used to index clusters within a batch of data
    """
    # Loop over the individual entries
    clusts, counts, single_counts, offsets = [], [], [], [0]
    for b in range(data.batch_size):
        # Get the list of clusters and cluster sizes within this entry
        data_b = data[b]
        clusts_b, counts_b = form_clusters(data_b, min_size, column, shapes)

        # Offset the cluster indexes appropriately
        for i, clust in enumerate(clusts_b):
            clusts_b[i] = clust + offsets[-1]

        # Append
        clusts.extend(clusts_b)
        counts.append(len(counts_b))
        single_counts.extend(counts_b)
        if b < (data.batch_size - 1):
            offsets.append(offsets[-1] + len(data_b))

    # Make an IndexBatch out of the list
    return IndexBatch(clusts, offsets, counts, single_counts, is_numpy=data.is_numpy)


def get_cluster_label_batch(data, clusts, column=CLUST_COL):
    """Batched version of :func:`get_cluster_label`.

    Parameters
    ----------
    data : TensorBatch
        Batch of cluster label data tensor
    clusts : IndexBatch
        (C) List of cluster indexes
    column : int, default CLUST_COL
        Column in the label tensor which specifies the requested label

    Returns
    -------
    TensorBatch
        (C) List of individual cluster labels
    """
    labels = get_cluster_label(data.tensor, clusts.index_list, column)

    return TensorBatch(labels, clusts.counts)


def get_cluster_closest_label_batch(data, coord_label, clusts, labels, default):
    """Batched version of :func:`get_cluster_closest_label`.

    Parameters
    ----------
    data : TensorBatch
        Batch of cluster label data tensor
    coord_label : TensorBatch
        Batch of particle end points labels
    clusts : IndexBatch
        (C) List of cluster indexes
    labels : TensorBatch
        (C) Existing list of cluster labels (the new labels will be a subset)
    default : Union[int, List[int]]
        Default value to assign to secondary clusters

    Returns
    -------
    TensorBatch
        (C) List of individual cluster labels
    """
    labels_adapt = np.empty(len(clusts.index_list), dtype=np.int64)
    for b in range(data.batch_size):
        lower, upper = clusts.edges[b], clusts.edges[b + 1]
        labels_adapt[lower:upper] = get_cluster_closest_label(
            data[b], coord_label[b], clusts[b], labels[b], default
        )

    return TensorBatch(labels_adapt, clusts.counts)


def get_cluster_primary_label_batch(data, clusts, column):
    """Batched version of :func:`get_cluster_primary_label`.

    Parameters
    ----------
    data : TensorBatch
        Batch of cluster label data tensor
    clusts : IndexBatch
        (C) List of cluster indexes
    column : int
        Column in the label tensor which specifies the requested label

    Returns
    -------
    TensorBatch
        (C) List of cluster primary labels
    """
    labels = get_cluster_primary_label(data.tensor, clusts.index_list, column)

    return TensorBatch(labels, clusts.counts)


def get_cluster_closest_primary_label_batch(data, coord_label, clusts, primary_ids):
    """Batched version of :func:`get_cluster_cloest_primary_label`.

    Parameters
    ----------
    data : TensorBatch
        Batch of cluster label data tensor
    coord_label : TensorBatch
        Batch of particle end points labels
    clusts : IndexBatch
        (C) List of cluster indexes
    primary_ids : TensorBatch
        (C) Existing list of primary IDs (the new labels will be a subset)

    Returns
    -------
    TensorBatch
        (C) List of cluster primary labels
    """
    labels = np.empty(len(clusts.index_list), dtype=np.int64)
    for b in range(data.batch_size):
        lower, upper = clusts.edges[b], clusts.edges[b + 1]
        labels[lower:upper] = get_cluster_closest_primary_label(
            data[b], coord_label[b], clusts[b], primary_ids[b]
        )

    return TensorBatch(labels, clusts.counts)


def get_cluster_points_label_batch(data, coord_label, clusts, random_order=True):
    """Batched version of :func:`get_cluster_points_label`

    Parameters
    ----------
    data : TensorBatch
        Batch of cluster label data tensor
    coord_label : TensorBatch
        Batch of particle end points labels
    clusts : IndexBatch
        (C) List of cluster indexes
    random_order : bool, default True
        If `True`, randomize the order in which the start en end points of
        a track are stored in the output

    Returns
    -------
    np.ndarray
        (C, 6) Cluster-wise start and end points (in random order if requested)
    """
    num_clusts = len(clusts.index_list)
    if isinstance(data.tensor, torch.Tensor):
        points = torch.empty((num_clusts, 6), dtype=data.dtype, device=data.device)
    else:
        points = np.empty((num_clusts, 6), dtype=data.dtype)

    for b in range(data.batch_size):
        lower, upper = clusts.edges[b], clusts.edges[b + 1]
        points[lower:upper] = get_cluster_points_label(
            data[b], coord_label[b], clusts[b], random_order
        )

    return TensorBatch(points, clusts.counts, coord_cols=points.shape[1])


def get_cluster_directions_batch(data, starts, clusts, max_dist=-1.0, optimize=False):
    """Batched version of :func:`get_cluster_directions`.

    Parameters
    ----------
    data : TensorBatch
        Batch of cluster label data tensor
    starts : TensorBatch
        (C, 3) Start points w.r.t. which to estimate the direction
    clusts : IndexBatch
        (C) List of cluster indexes
    max_dist : float, default -1.0
        Neighborhood radius around the point used to estimate the direction
    optimize : bool, default False
        If `True`, the neighborhood radius is optimized on the fly for
        each cluster.

    Returns
    -------
    TensorBatch
        (C, 3) List of cluster directions
    """
    dirs = get_cluster_directions(
        data.tensor, starts.tensor, clusts.index_list, max_dist, optimize
    )

    return TensorBatch(dirs, clusts.counts)


def get_cluster_dedxs_batch(data, starts, clusts, max_dist=-1.0):
    """Batched version of :func:`get_cluster_dedxs`.

    Parameters
    ----------
    data : TensorBatch
        Batch of cluster label data tensor
    starts : TensorBatch
        (C, 3) Start points w.r.t. which to estimate the direction
    clusts : IndexBatch
        (C) List of cluster indexes
    max_dist : float, default -1.0
        Neighborhood radius around the point used t compute the dE/dx

    Returns
    -------
    TensorBatch
        (C) List of cluster dE/dx value close to the start points
    """
    dedxs = get_cluster_dedxs(data.tensor, starts.tensor, clusts.index_list, max_dist)

    return TensorBatch(dedxs, clusts.counts)


def get_cluster_features_batch(data, clusts, add_value=False, add_shape=False):
    """Batched version of :func:`get_cluster_features`.

    Parameters
    ----------
    data : TensorBatch
        Batch of cluster label data tensor
    starts : TensorBatch
        (C, 3) Start points w.r.t. which to estimate the direction
    clusts : IndexBatch
        (C) List of cluster indexes
    max_dist : float, default -1.0
        Neighborhood radius around the point used t compute the dE/dx

    Returns
    -------
    TensorBatch
        (C) List of cluster dE/dx value close to the start points
    """
    feats = get_cluster_features(data.tensor, clusts.index_list, add_value, add_shape)

    return TensorBatch(feats, clusts.counts)


def form_clusters(data, min_size=-1, column=CLUST_COL, shapes=None):
    """Builds a list of indexes corresponding to each cluster in the event.

    The `data` tensor should only contain one entry.

    Parameters
    ----------
    data : Union[np.ndarray, torch.Tensor]
        Cluster label data tensor
    min_size : int, default -1
        Minimum size of a cluster to be included in the list
    column : int, default CLUST_COL
        Column of the label tensor to use to fetch the pixel cluster IDs
    shapes : List[int], optional
        List of semantic classes to include in the list of cluster

    Returns
    -------
    List[Union[np.ndarray, torch.Tensor]]
        (C) List of arrays of voxel indexes in each cluster
    np.ndarray
        (C) Number of pixels in the mask for each cluster
    """
    # Dispatch to the right functions based on input type
    if isinstance(data, torch.Tensor):
        return _form_clusters_torch(data, min_size, column, shapes)
    else:
        return _form_clusters_np(data, min_size, column, shapes)


def _form_clusters_torch(data, min_size, column, shapes):
    # If requested, restrict data to a specific set of semantic classes
    shape_index = None
    if shapes is not None:
        shapes = torch.as_tensor(shapes, dtype=data.dtype, device=data.device)
        shape_index = torch.where(torch.any(data[:, SHAPE_COL] == shapes[:, None], 0))[
            0
        ]
        data = data[shape_index]

    # Get the list of unique clusters in this entry, order indices
    clust_ids = data[:, column]
    uniques, counts = torch.unique(clust_ids, return_counts=True)
    full_index = torch.argsort(clust_ids, stable=True)
    if shape_index is not None:
        full_index = shape_index[full_index]

    # Build valid index
    valid_index = (
        torch.where((counts >= min_size) & (uniques > -1))[0].detach().cpu().numpy()
    )

    # Build index list, restrict to valid clusters
    counts = counts.detach().cpu().numpy()
    breaks = tuple(np.cumsum(counts)[:-1])
    clusts = torch.tensor_split(full_index, breaks)

    # Restrict to valid clusters
    clusts = [clusts[i] for i in valid_index]
    counts = counts[valid_index]

    return clusts, counts


def _form_clusters_np(data, min_size, column, shapes):
    # If requested, restrict data to a specific set of semantic classes
    shape_index = None
    if shapes is not None:
        shapes = np.array(shapes, dtype=data.dtype)
        shape_index = np.where(np.any(data[:, SHAPE_COL] == shapes[:, None], 0))[0]
        data = data[shape_index]

    # Get the clusters in this entry
    clust_ids = data[:, column]
    uniques, counts = np.unique(clust_ids, return_counts=True)
    full_index = np.argsort(clust_ids, stable=True)
    if shape_index is not None:
        full_index = shape_index[full_index]

    # Build valid index
    valid_index = np.where((counts >= min_size) & (uniques > -1))[0]

    # Build index list, restrict to valid clusters
    breaks = tuple(np.cumsum(counts)[:-1])
    clusts = list(np.split(full_index, breaks))

    # Restrict to valid clusters
    clusts = [clusts[i] for i in valid_index]
    counts = counts[valid_index]

    return clusts, counts


@numbafy(cast_args=["data"], list_args=["clusts"], keep_torch=True, ref_arg="data")
def break_clusters(data, clusts, eps, metric_id, p):
    """Runs DBSCAN on each invididual cluster to segment them further if needed.

    Parameters
    ----------
    data : np.ndarray
        Cluster label data tensor
    clusts : List[np.ndarray]
        (C) List of cluster indexes
    eps : float
        DBSCAN clustering distance scale
    metric_id : int
        DBSCAN clustering distance metric enumerator
    p : float
        p-norm factor for the Minkowski metric, if used

    Returns
    -------
    np.ndarray
        New array of broken cluster labels
    """
    if len(clusts) == 0:
        return np.copy(data[:, CLUST_COL])

    # Break labels
    break_labels = _break_clusters(data, clusts, eps, metric_id, p)

    # Offset individual broken labels to prevent overlap
    labels = np.copy(data[:, CLUST_COL])
    offset = np.max(labels) + 1
    for clust in clusts:
        # Update IDs, offset
        ids = break_labels[clust]
        labels[clust] = offset + ids
        offset += len(np.unique(ids))

    return labels


@nb.njit(cache=True, parallel=True, nogil=True)
def _break_clusters(
    data: nb.float32[:, :],
    clusts: nb.types.List(nb.int64[:]),
    eps: nb.float64,
    metric_id: nb.int64,
    p: nb.float64,
) -> nb.int64[:]:
    # Loop over clusters to break, run DBSCAN
    break_labels = np.full(len(data), -1, dtype=data.dtype)
    points = data[:, COORD_COLS]
    for k in nb.prange(len(clusts)):
        # Restrict the points to those in the cluster
        clust = clusts[k]
        points_c = points[clust]

        # Run DBSCAN on the cluster, update labels
        clust_ids = sm.cluster.dbscan(points_c, eps=eps, metric_id=metric_id, p=p)

        # Store the breaking IDs
        break_labels[clust] = clust_ids

    return break_labels


@numbafy(cast_args=["data"], list_args=["clusts"])
def get_cluster_label(data, clusts, column=CLUST_COL):
    """Returns the majority label of each cluster, specified by the
    requested data column of the label tensor.

    Parameters
    ----------
    data : np.ndarray
        Cluster label data tensor
    clusts : List[np.ndarray]
        (C) List of cluster indexes
    column : int, default CLUST_COL
        Column in the label tensor which specifies the requested label

    Returns
    -------
    np.ndarray
        (C) List of individual cluster labels
    """
    if len(clusts) == 0:
        return np.empty(0, dtype=data.dtype)

    return _get_cluster_label(data, clusts, column)


@nb.njit(cache=True)
def _get_cluster_label(
    data: nb.float64[:, :],
    clusts: nb.types.List(nb.int64[:]),
    column: nb.int64 = CLUST_COL,
) -> nb.float64[:]:

    labels = np.empty(len(clusts), dtype=data.dtype)
    for i, c in enumerate(clusts):
        v, cts = sm.unique(data[c, column])
        labels[i] = v[np.argmax(cts)]

    return labels


@numbafy(cast_args=["data", "coord_label"], list_args=["clusts"])
def get_cluster_closest_label(data, coord_label, clusts, labels, default):
    """Sets the label of clusters based on their proximity to the start point
    of the particle which created them.

    Parameters
    ----------
    data : np.ndarray
        Cluster label data tensor
    coord_label : np.ndarray
        Coordinate labels associated with each particle
    clusts : List[np.ndarray]
        (C) List of cluster indexes
    labels : np.ndarray
        (C) Existing list of cluster labels (the new labels will be a subset)
    default : np.ndarray
        Default label to assign to secondary clusters

    Returns
    -------
    np.ndarray
        (C) List of cluster labels
    """
    if len(clusts) == 0:
        return np.empty(0, dtype=data.dtype)

    return _get_cluster_closest_label(data, coord_label, clusts, labels, default)


@nb.njit(cache=True)
def _get_cluster_closest_label(
    data: nb.float64[:, :],
    coord_label: nb.float64[:, :],
    clusts: nb.types.List(nb.int64[:]),
    labels: nb.float64[:],
    default: nb.int64[:],
) -> nb.float64[:]:

    # Loop over the unique cluster groups
    group_ids = _get_cluster_label(data, clusts, GROUP_COL)
    labels_adapt = labels.copy()
    voxels = data[:, COORD_COLS]
    points = coord_label[:, COORD_COLS]
    for g in np.unique(group_ids.astype(np.int64)):
        # If the group index does not exist in the points, do not touch labels
        group_index = np.where(group_ids == g)[0]
        if g < 0 or g >= len(points):
            continue

        # Get the coordinates of the start point
        start_point = points[g].reshape(-1, 3)

        # Minimize the point-cluster distances
        dists = np.empty(len(group_index), dtype=data.dtype)
        for i, c in enumerate(group_index):
            dists[i] = np.min(sm.distance.cdist(start_point, voxels[clusts[c]]))

        # Label the closest cluster as the original label only, assign default
        # values ot the other clusters in the group
        closest_index = group_index[np.argmin(dists)]
        label = int(labels[closest_index])
        deflt = default[label] if label > -1 and label < len(default) else -1

        labels_adapt[group_index] = deflt
        labels_adapt[closest_index] = label

    return labels_adapt


@numbafy(cast_args=["data"], list_args=["clusts"])
def get_cluster_primary_label(data, clusts, column):
    """Returns the majority label of the primary cluster of the group each
    cluster belongs to, specified in the requested data column of the label
    tensor.

    The primary component is identified by picking the set of label voxels
    that have a `PART_COL` identical to the cluster `GROUP_COL`.

    Parameters
    ----------
    data : np.ndarray
        Cluster label data tensor
    clusts : List[np.ndarray]
        (C) List of cluster indexes
    column : int
        Column in the label tensor which specifies the requested label

    Returns
    -------
    np.ndarray
        (C) List of cluster primary labels
    """
    if len(clusts) == 0:
        return np.empty(0, dtype=data.dtype)

    return _get_cluster_primary_label(data, clusts, column)


@nb.njit(cache=True)
def _get_cluster_primary_label(
    data: nb.float64[:, :], clusts: nb.types.List(nb.int64[:]), column: nb.int64
) -> nb.float64[:]:

    labels = np.empty(len(clusts), dtype=data.dtype)
    group_ids = _get_cluster_label(data, clusts, GROUP_COL)
    for i, clust in enumerate(clusts):
        part_ids = data[clust, PART_COL]
        primary_mask = np.where(part_ids == group_ids[i])[0]
        if len(primary_mask):
            # Only use the primary component to label the cluster
            v, cts = sm.unique(data[clust[primary_mask], column])
        else:
            # If there is no primary contribution, use the whole cluster
            v, cts = sm.unique(data[clust, column])
        labels[i] = v[np.argmax(cts)]

    return labels


@numbafy(cast_args=["data", "coord_label"], list_args=["clusts"])
def get_cluster_closest_primary_label(data, coord_label, clusts, primary_ids):
    """Sets the primary label of clusters based on their proximity to the start
    point of the particle which created them.

    Parameters
    ----------
    data : np.ndarray
        Cluster label data tensor
    coord_label : np.ndarray
        Coordinate labels associated with each particle
    clusts : List[np.ndarray]
        (C) List of cluster indexes
    primary_ids : np.ndarray
        (C) Existing list of primary IDs (the new labels will be a subset)

    Returns
    -------
    np.ndarray
        (C) List of cluster primary labels
    """
    if len(clusts) == 0:
        return np.empty(0, dtype=data.dtype)

    return _get_cluster_closest_primary_label(data, coord_label, clusts, primary_ids)


@nb.njit(cache=True)
def _get_cluster_closest_primary_label(
    data: nb.float64[:, :],
    coord_label: nb.float64[:, :],
    clusts: nb.types.List(nb.int64[:]),
    primary_ids: nb.float64[:],
) -> nb.float64[:]:

    # Loop over the unique primary cluster groups
    primary_index = np.where(primary_ids == 1)[0]
    group_ids = _get_cluster_label(data, clusts, GROUP_COL)[primary_index]
    labels = primary_ids.copy()
    voxels = data[:, COORD_COLS]
    points = coord_label[:, COORD_COLS]
    for g in np.unique(group_ids.astype(np.int64)):
        # If the group index does not exist in the points, do not touch labels
        group_index = primary_index[group_ids == g]
        if g < 0 or g >= len(points):
            continue

        # Get the coordinates of the start point
        start_point = points[g].reshape(-1, 3)

        # Minimize the point-cluster distances
        dists = np.empty(len(group_index), dtype=data.dtype)
        for i, c in enumerate(group_index):
            dists[i] = np.min(sm.distance.cdist(start_point, voxels[clusts[c]]))

        # Label the closest cluster as the only primary cluster
        labels[group_index] = 0
        labels[group_index[np.argmin(dists)]] = 1

    return labels


@numbafy(cast_args=["data"], list_args=["clusts"], keep_torch=True, ref_arg="data")
def get_cluster_centers(data, clusts):
    """Returns the coordinate of the centroid associated with each cluster.

    Parameters
    ----------
    data : np.ndarray
        Cluster label data tensor
    clusts : List[np.ndarray]
        (C) List of cluster indexes

    Returns
    -------
    np.ndarray
        (C, 3) Tensor of cluster centers
    """
    if len(clusts) == 0:
        return np.empty((0, 3), dtype=data.dtype)

    return _get_cluster_centers(data, clusts)


@nb.njit(cache=True)
def _get_cluster_centers(
    data: nb.float64[:, :], clusts: nb.types.List(nb.int64[:])
) -> nb.float64[:, :]:

    centers = np.empty((len(clusts), 3), dtype=data.dtype)
    for i, c in enumerate(clusts):
        centers[i] = np.sum(data[c][:, COORD_COLS], axis=0) / len(c)

    return centers


@numbafy(cast_args=["data"], list_args=["clusts"])
def get_cluster_sizes(data, clusts):
    """Returns the sizes of each cluster.

    Parameters
    ----------
    data : np.ndarray
        Cluster label data tensor
    clusts : List[np.ndarray]
        (C) List of cluster indexes

    Returns
    -------
    np.ndarray
        (C) List of cluster sizes
    """
    if len(clusts) == 0:
        return np.empty(0, dtype=np.int64)

    return _get_cluster_sizes(data, clusts)


@nb.njit(cache=True)
def _get_cluster_sizes(
    data: nb.float64[:, :], clusts: nb.types.List(nb.int64[:])
) -> nb.int64[:]:

    sizes = np.empty(len(clusts), dtype=np.int64)
    for i, c in enumerate(clusts):
        sizes[i] = len(c)

    return sizes


@numbafy(cast_args=["data"], list_args=["clusts"], keep_torch=True, ref_arg="data")
def get_cluster_energies(data, clusts):
    """Returns the total charge/energy deposited by each cluster.

    Parameters
    ----------
    data : np.ndarray
        Cluster label data tensor
    clusts : List[np.ndarray]
        (C) List of cluster indexes

    Returns
    -------
    np.ndarray
        (C) List of cluster pixel sums
    """
    if len(clusts) == 0:
        return np.empty(0, dtype=data.dtype)

    return _get_cluster_energies(data, clusts)


@nb.njit(cache=True)
def _get_cluster_energies(
    data: nb.float64[:, :], clusts: nb.types.List(nb.int64[:])
) -> nb.float64[:]:

    energies = np.empty(len(clusts), dtype=data.dtype)
    for i, c in enumerate(clusts):
        energies[i] = np.sum(data[c, VALUE_COL])

    return energies


def get_cluster_features(data, clusts, add_value=False, add_shape=False):
    """Returns an array of features for each cluster.

    The basic 16 geometric features are composed of:
    - Center (3)
    - Covariance matrix (9)
    - Principal axis (3)
    - Voxel count (1)

    The flag `add_value` adds the following 2 features:
    - Mean energy (1)
    - RMS energy (1)

    The flag `add_shape` adds the particle shape information:
    - Semantic type (1), i.e. most represented type in cluster

    Parameters
    ----------
    data : np.ndarray
        Cluster label data tensor
    clusts : List[np.ndarray]
        (C) List of cluster indexes

    Returns
    -------
    np.ndarray
        (C, N_c) Tensor of cluster features
    """
    feats = get_cluster_features_base(data, clusts)
    if add_value or add_shape:
        feats_ext = get_cluster_features_extended(data, clusts, add_value, add_shape)
        if isinstance(data, np.ndarray):
            feats = np.hstack((feats, feats_ext))
        else:
            feats = torch.cat((feats, feats_ext), dim=1)

    return feats


@numbafy(cast_args=["data"], list_args=["clusts"], keep_torch=True, ref_arg="data")
def get_cluster_features_base(data, clusts):
    """Returns an array of 16 geometric features for each of cluster.

    The 16 geometric features are composed of:
    - Center (3)
    - Covariance matrix (9)
    - Principal axis (3)
    - Voxel count (1)

    Parameters
    ----------
    data : np.ndarray
        Cluster label data tensor
    clusts : List[np.ndarray]
        (C) List of cluster indexes

    Returns
    -------
    np.ndarray
        (C, 16) Tensor of cluster features
    """
    if len(clusts) == 0:
        return np.empty((0, 16), dtype=data.dtype)  # Cannot type empty list

    return _get_cluster_features_base(data, clusts)


@nb.njit(parallel=True, cache=True)
def _get_cluster_features_base(
    data: nb.float64[:, :], clusts: nb.types.List(nb.int64[:])
) -> nb.float64[:, :]:

    # Loop over the clusters (parallelize). The `prange` function creates a
    # uint64 iterator which is cast to int64 to access a list, and throws a
    # warning. To avoid this, use a separate counter to acces clusts.
    feats = np.empty((len(clusts), 16), dtype=data.dtype)
    ids = np.arange(len(clusts)).astype(np.int64)
    for k in nb.prange(len(clusts)):
        # Get list of voxels in the cluster
        clust = clusts[ids[k]]
        x = data[clust][:, COORD_COLS]

        # Get cluster center
        center = sm.mean(x, 0)

        # Get orientation matrix
        A = np.cov(x.T, ddof=len(x) - 1).astype(x.dtype)

        # Center data
        x = x - center

        # Get eigenvectors, normalize orientation matrix and eigenvalues to
        # largest. If points are superimposed, i.e. if the largest eigenvalue
        # != 0, no need to keep going
        # TODO: get rid of casting, this is a complex LAPACK issue currently
        w, v = np.linalg.eigh(A.astype(np.float64))
        w, v = w.astype(x.dtype), v.astype(x.dtype)
        if w[2] == 0.0:
            feats[k] = np.concatenate((center, np.zeros(12), np.array([len(clust)])))
            continue
        dirwt = 1.0 - w[1] / w[2]
        B = A / w[2]

        # Get the principal direction, identify the direction of the spread
        v0 = v[:, 2]

        # Projection all points, x, along the principal axis
        x0 = np.dot(x, v0)

        # Evaluate the distance from the points to the principal axis
        xp0 = x - np.outer(x0, v0)
        np0 = np.empty(len(xp0), dtype=data.dtype)
        for i in range(len(xp0)):
            np0[i] = np.linalg.norm(xp0[i])

        # Flip the principal direction if it is not pointing towards the
        # maximum spread
        sc = np.dot(x0, np0)
        if sc < 0:
            # Numba does not support unary `-`, have to flip manually
            v0 = np.zeros(3, dtype=data.dtype) - v0

        # Weight direction
        v0 = dirwt * v0

        # Append
        feats[k] = np.concatenate((center, B.flatten(), v0, np.array([len(clust)])))

    return feats


@numbafy(cast_args=["data"], list_args=["clusts"], keep_torch=True, ref_arg="data")
def get_cluster_features_extended(data, clusts, add_value=True, add_shape=True):
    """Returns an array of 3 additional features for each of cluster.

    The flag `add_value` adds the following 2 features:
    - Mean energy (1)
    - RMS energy (1)

    The flag `add_shape` adds the particle shape information:
    - Semantic type (1), i.e. most represented type in cluster

    Parameters
    ----------
    data : np.ndarray
        Cluster label data tensor
    clusts : List[np.ndarray]
        (C) List of cluster indexes
    add_value : bool, default True
        Whether to add the mean and std of the pixel values
    add_shape : bool, default True
        Whether to add the shape of the cluster

    Returns
    -------
    np.ndarray
        (C, 1/2/3) Tensor of additional cluster features
    """
    assert (
        add_value or add_shape
    ), "Must add either value or shape for this function to do anything"
    if len(clusts) == 0:
        return np.empty((0, add_value * 2 + add_shape), dtype=data.dtype)

    return _get_cluster_features_extended(data, clusts, add_value, add_shape)


@nb.njit(parallel=True, cache=True)
def _get_cluster_features_extended(
    data: nb.float64[:, :],
    clusts: nb.types.List(nb.int64[:]),
    add_value: bool = True,
    add_shape: bool = True,
) -> nb.float64[:, :]:
    feats = np.empty((len(clusts), add_value * 2 + add_shape), dtype=data.dtype)
    ids = np.arange(len(clusts)).astype(np.int64)
    for k in nb.prange(len(clusts)):
        # Get cluster
        clust = clusts[ids[k]]

        # Get mean and RMS energy in the cluster, if requested
        if add_value:
            mean_value = np.mean(data[clust, VALUE_COL])
            std_value = np.std(data[clust, VALUE_COL])
            feats[k, :2] = np.array([mean_value, std_value], dtype=data.dtype)

        # Get the cluster semantic class, if requested
        if add_shape:
            types, cnts = sm.unique(data[clust, SHAPE_COL])
            major_sem_type = types[np.argmax(cnts)]
            feats[k, -1] = major_sem_type

    return feats


@numbafy(
    cast_args=["data", "coord_label"],
    list_args=["clusts"],
    keep_torch=True,
    ref_arg="data",
)
def get_cluster_points_label(data, coord_label, clusts, random_order=True):
    """Gets label points for each cluster.

    Returns start point of primary shower fragment twice if shower, delta or
    Michel and both end points of tracks if track.

    Parameters
    ----------
    data : np.ndarray
        Cluster label data tensor
    coord_label : np.ndarray
        (P, 9) Particle end points labels
        [batch_id, start_x, start_y, start_z, end_x, end_y, end_z, time, shape]
    clusts : List[np.ndarray]
        (C) List of cluster indexes
    random_order : bool, default True
        If `True`, randomize the order in which the start en end points of
        a track are stored in the output

    Returns
    -------
    np.ndarray
        (C, 6) Cluster start and end points (in random order if requested)
    """
    if len(clusts) == 0:
        return np.empty((0, 6), dtype=data.dtype)

    return _get_cluster_points_label(data, coord_label, clusts, random_order)


@nb.njit(cache=True)
def _get_cluster_points_label(
    data: nb.float64[:, :],
    coord_label: nb.float64[:, :],
    clusts: nb.types.List(nb.int64[:]),
    random_order: nb.boolean = True,
) -> nb.float64[:, :]:

    # Get start and end points (one and the same for all but track class)
    points = np.empty((len(clusts), 6), dtype=data.dtype)
    for i, c in enumerate(clusts):
        # Use the first cluster in time
        part_ids = np.unique(data[c, PART_COL]).astype(np.int64)
        min_id = part_ids[np.argmin(coord_label[part_ids, COORD_TIME_COL])]
        min_label = coord_label[min_id]
        start, end = min_label[COORD_START_COLS], min_label[COORD_END_COLS]
        if random_order and np.random.choice(2):
            start, end = end, start

        points[i, :3] = start
        points[i, 3:6] = end

    # Bring the start points to the closest point in the corresponding cluster
    for i, c in enumerate(clusts):
        dist_mat = sm.distance.cdist(points[i].reshape(-1, 3), data[c][:, COORD_COLS])
        argmins = sm.argmin(dist_mat, axis=1)
        points[i] = data[c][argmins][:, COORD_COLS].reshape(-1)

    return points


@numbafy(
    cast_args=["data", "starts"], list_args=["clusts"], keep_torch=True, ref_arg="data"
)
def get_cluster_directions(data, starts, clusts, max_dist=-1.0, optimize=False):
    """Estimates the direction of each cluster.

    Parameters
    ----------
    data : np.ndarray
        Cluster label data tensor
    starts : np.ndarray
        (C, 3) Start points w.r.t. which to estimate the direction
    clusts : List[np.ndarray]
        (C) List of cluster indexes
    max_dist : float, default -1.0
        Neighborhood radius around the point used to estimate the direction
    optimize : bool, default False
        If `True`, the neighborhood radius is optimized on the fly for
        each cluster.

    Returns
    -------
    torch.tensor:
        (C, 3) Direction vector of each cluster
    """
    if len(clusts) == 0:
        return np.empty(starts.shape, dtype=data.dtype)
    if data.shape[1] > 3:
        data = data[:, COORD_COLS]

    return _get_cluster_directions(data, starts, clusts, max_dist, optimize)


@nb.njit(parallel=True, cache=True)
def _get_cluster_directions(
    voxels: nb.float64[:, :],
    starts: nb.float64[:, :],
    clusts: nb.types.List(nb.int64[:]),
    max_dist: nb.float64 = -1.0,
    optimize: nb.boolean = False,
) -> nb.float64[:, :]:

    dirs = np.empty(starts.shape, voxels.dtype)
    ids = np.arange(len(clusts)).astype(np.int64)
    for k in nb.prange(len(clusts)):
        dirs[k] = cluster_direction(
            voxels[clusts[ids[k]]], starts[k], max_dist, optimize
        )

    return dirs


@nb.njit(cache=True)
def cluster_direction(
    voxels: nb.float64[:, :],
    start: nb.float64[:],
    max_dist: nb.float64 = -1,
    optimize: nb.boolean = False,
) -> nb.float64[:]:
    """Estimates the orientation of a cluster.

    It follows the following procedure:
    - By default, it takes the normalized mean direction from the cluster
      start point to the cluster voxels
    - If `max_dist` is specified, it restricts the cluster voxels
      to those within a `max_dist` radius from the start point
    - If `optimize` is True, it selects the neighborhood which
      minimizes the transverse spread w.r.t. the direction

    Parameters
    ----------
    voxels : np.ndarray
        (N, 3) Voxel coordinates
    starts : np.ndarray
        (C, 3) Start points w.r.t. which to estimate the direction
    clusts : List[np.ndarray]
        (C) List of cluster indexes
    max_dist : float, default -1.0
        Neighborhood radius around the point used to estimate the direction
    optimize : bool, default False
        If `True`, the neighborhood radius is optimized on the fly for
        each cluster.

    Returns
    -------
    np.ndarray
        (3) Direction vector
    """
    # If max_dist is set, limit the set of voxels to those within a sphere
    # of radius max_dist
    assert (
        voxels.shape[1] == 3
    ), "The shape of the input is not compatible with voxel coordinates."

    if max_dist > 0:
        dist_mat = sm.distance.cdist(start.reshape(1, -1), voxels).flatten()
        voxels = voxels[dist_mat <= max(max_dist, np.min(dist_mat))]

    # If optimize is set, select the radius by minimizing the transverse spread
    if optimize and len(voxels) > 2:
        # Order the cluster points by increasing distance to the start point
        dist_mat = sm.distance.cdist(start.reshape(1, -1), voxels).flatten()
        order = np.argsort(dist_mat)
        voxels = voxels[order]
        dist_mat = dist_mat[order]

        # Find the PCA relative secondary spread for each point
        labels = -np.ones(len(voxels), dtype=voxels.dtype)
        meank = sm.mean(voxels[:3], 0)
        covk = (np.transpose(voxels[:3] - meank) @ (voxels[:3] - meank)) / 3
        for i in range(2, len(voxels)):
            # Get the eigenvalues, compute relative transverse spread
            # TODO: get rid of casting, this is complex LAPACK issue currently
            w = np.linalg.eigvalsh(covk.astype(np.float64)).astype(voxels.dtype)
            labels[i] = (
                np.sqrt(w[2] / (w[0] + w[1])) if (w[0] + w[1]) / w[2] > 1e-6 else 0.0
            )

            # If the value is the same as the previous, choose this one
            if labels[i] == labels[i - 1]:
                labels[i - 1] = -1.0

            # Increment mean and matrix
            if i != len(voxels) - 1:
                meank = ((i + 1) * meank + voxels[i + 1]) / (i + 2)
                covk = (i + 1) * covk / (i + 2) + (voxels[i + 1] - meank).reshape(
                    -1, 1
                ) * (voxels[i + 1] - meank) / (i + 1)

        # Subselect voxels that are most track-like
        max_id = np.argmax(labels)
        voxels = voxels[: max_id + 1]

    # If no voxels were selected, return dummy value
    if len(voxels) == 0 or (len(voxels) == 1 and np.all(voxels[0] == start)):
        return np.array([1.0, 0.0, 0.0], dtype=voxels.dtype)

    # Compute mean direction with respect to start point, normalize it
    rel_voxels = np.empty((len(voxels), 3), dtype=voxels.dtype)
    for i in range(len(voxels)):
        rel_voxels[i] = voxels[i] - start

    mean = sm.mean(rel_voxels, 0)
    norm = np.sqrt(np.sum(mean**2))
    if norm:
        return mean / norm

    return mean


@numbafy(
    cast_args=["data", "starts"], list_args=["clusts"], keep_torch=True, ref_arg="data"
)
def get_cluster_dedxs(data, starts, clusts, max_dist=-1.0, anchor=False):
    """Computes the initial local dE/dxs of each cluster.

    Parameters
    ----------
    data : np.ndarray
        Cluster label data tensor
    starts : np.ndarray
        (C, 3) Start points w.r.t. which to estimate the local dE/dxs
    clusts : List[np.ndarray]
        (C) List of cluster indexes
    max_dist : float, default -1.0
        Neighborhood radius around the point used to compute the dE/dx
    anchor : bool, default False
        If true, anchor the start point to the closest cluster point

    Returns
    -------
    np.ndarray
        (C) Local dE/dx values for each cluster
    """
    if len(clusts) == 0:
        return np.empty(0, dtype=data.dtype)

    return _get_cluster_dedxs(
        data[:, COORD_COLS], data[:, VALUE_COL], starts, clusts, max_dist, anchor
    )


@nb.njit(parallel=True, cache=True)
def _get_cluster_dedxs(
    voxels: nb.float64[:, :],
    values: nb.float64[:],
    starts: nb.float64[:, :],
    clusts: nb.types.List(nb.int64[:]),
    max_dist: nb.float64 = -1,
    anchor: nb.boolean = False,
) -> nb.float64[:, :]:

    dedxs = np.empty(len(clusts), voxels.dtype)
    ids = np.arange(len(clusts)).astype(np.int64)
    for k in nb.prange(len(clusts)):
        dedxs[k] = cluster_dedx(
            voxels[clusts[ids[k]]],
            values[clusts[ids[k]]],
            starts[k].astype(np.float64),
            max_dist,
            anchor,
        )

    return dedxs


@nb.njit(cache=True)
def cluster_dedx(
    voxels: nb.float64[:, :],
    values: nb.float64[:],
    start: nb.float64[:],
    max_dist: nb.float64 = 5.0,
    anchor: nb.boolean = False,
) -> nb.float64[:]:
    """Computes the initial local dE/dx of a cluster.

    Parameters
    ----------
    voxels : np.ndarray
        (N, 3) Voxel coordinates
    values : np.ndarray
        (N) Voxel values
    start : np.ndarray
        (3) Start point w.r.t. which to compute the local dE/dx
    max_dist : float, default 5.0
        Neighborhood radius around the point used to compute the dE/dx
    anchor : bool, default False
        If true, anchor the start point to the closest cluster point

    Returns
    -------
    float
        Local dE/dx value around the start point
    """
    # Sanity check
    assert (
        voxels.shape[1] == 3
    ), "The shape of the input is not compatible with voxel coordinates."

    # If necessary, anchor start point to the closest cluster point
    if anchor:
        dists = sm.distance.cdist(start.reshape(1, -1), voxels).flatten()
        start = voxels[np.argmin(dists)].astype(start.dtype)  # Dirty

    # If max_dist is set, limit the set of voxels to those within a sphere of
    # radius max_dist around the start point
    dists = sm.distance.cdist(start.reshape(1, -1), voxels).flatten()
    if max_dist > 0.0:
        index = np.where(dists <= max_dist)[0]
        if len(index) < 2:
            return 0.0

        values, dists = values[index], dists[index]

    # Compute the total energy in the neighborhood and the max distance, return ratio
    if np.max(dists) == 0.0:
        return 0.0

    return np.sum(values) / np.max(dists)


@nb.njit(cache=True)
def cluster_dedx_dir(
    voxels: nb.float64[:, :],
    values: nb.float64[:],
    start: nb.float64[:],
    start_dir: nb.float64[:],
    max_dist: nb.float64 = 3.0,
    anchor: nb.boolean = False,
) -> nb.float64[:]:
    """Computes the initial local dE/dx of a cluster by leveraging an
    existing cluster direction estimate.

    Parameters
    ----------
    voxels : np.ndarray
        (N, 3) Voxel coordinates
    values : np.ndarray
        (N) Voxel values
    start : np.ndarray
        (3) Start point w.r.t. which to compute the local dE/dx
    start_dir : np.ndarray
        (3) Start direction of the cluster
    max_dist : float, default 5.0
        Neighborhood radius around the point used to compute the dE/dx

    Returns
    -------
    float
        Local dE/dx value around the start point
    float
        Energy deposited around the start point (dE)
    float
        Length around the start point (dx)
    float
        Relative spread around the cluster direction (quality metric)
    int
        Number of voxels in the neighborhood around the start poin
    """
    # Sanity check
    assert (
        voxels.shape[1] == 3
    ), "The shape of the input is not compatible with voxel coordinates."

    # If necessary, anchor start point to the closest cluster point
    if anchor:
        dists = sm.distance.cdist(start.reshape(1, -1), voxels).flatten()
        start = voxels[np.argmin(dists)].astype(start.dtype)  # Dirty

    # If max_dist is set, limit the set of voxels to those within a sphere of
    # radius max_dist around the start point
    dists = sm.distance.cdist(start.reshape(1, -1), voxels).flatten()
    if max_dist > 0.0:
        index = np.where(dists <= max_dist)[0]
        if len(index) < 2:
            return 0.0, 0.0, 0.0, -1.0, len(index)

        voxels, values, dists = voxels[index], values[index], dists[index]

    # Project the voxels within the sphere onto the reco direction
    voxels_proj = np.dot(voxels - start, start_dir)

    # Mask the voxels to only include the top hemisphere
    index = np.where((voxels_proj >= -1e-3) & (voxels_proj <= max_dist))[0]
    if len(index) < 2:
        return 0.0, 0.0, 0.0, -1.0, len(index)

    voxels, voxels_proj, values = voxels[index], voxels_proj[index], values[index]

    # Compute length as the length along the start direction
    dE = np.sum(values)
    dx = np.max(voxels_proj) - np.min(voxels_proj)

    # Calculate the mean spread from the reco direction (quality metric)
    voxels_sp = voxels - start
    vectors_to_axis = voxels_sp - np.outer(voxels_proj, start_dir)
    spreads = np.sqrt(np.sum(vectors_to_axis**2, axis=1))
    spread = np.sum(spreads) / len(index)

    return dE / dx, dE, dx, spread, len(index)


@numbafy(cast_args=["data"], list_args=["clusts"], keep_torch=True, ref_arg="data")
def get_cluster_start_points(data, clusts):
    """Estimates the start point of clusters based on their PCA and the
    local curvature at each of the PCA extrema.

    Parameters
    ----------
    data : np.ndarray
        Cluster label data tensor
    clusts : List[np.ndarray]
        (C) List of cluster indexes

    Returns
    -------
    np.ndarray
        (C, 3) Cluster start points
    """
    if len(clusts) == 0:
        return np.empty((0, 3), dtype=data.dtype)

    return _get_cluster_start_points(data, clusts)


@nb.njit(parallel=True, cache=True)
def _get_cluster_start_points(
    data: nb.float64[:, :], clusts: nb.types.List(nb.int64[:])
) -> nb.float64[:, :]:

    points = np.empty((len(clusts), 3))
    for k in nb.prange(len(clusts)):
        vid = cluster_end_points(data[clusts[k]][:, COORD_COLS])[-1]

    return points


@nb.njit(cache=True)
def cluster_end_points(voxels: nb.float64[:, :]) -> (nb.float64[:], nb.float64[:]):
    """Estimates the end points of a clusters using PCA and curvature.

    It proceeds in the following fashion:
    1. Find the principal axis a of the point cloud
    2. Find the coordinate a_i of each point along this axis
    3. Find the points with minimum and maximum coordinate
    4. Find the point that has the largest umbrella curvature

    Parameters
    ----------
    voxels : np.ndarray
        (N, 3) Voxel coordinates

    Returns
    -------
    int
        Index of the start voxel
    int
        Index of the end voxel
    """
    # Get the axis of maximum spread
    axis = sm.decomposition.principal_components(voxels)[0]

    # Compute coord values along that axis
    coords = np.empty(len(voxels))
    for i in range(len(coords)):
        coords[i] = np.dot(voxels[i], axis)

    # Compute curvature of the extremities
    ids = [np.argmin(coords), np.argmax(coords)]

    # Sort the voxel IDs by increasing order of curvature order
    curvs = [umbrella_curv(voxels, ids[0]), umbrella_curv(voxels, ids[1])]
    curvs = np.array(curvs, dtype=np.int64)
    ids = np.array(ids, dtype=np.int64)
    ids[np.argsort(curvs)]

    # Return extrema
    return voxels[ids[0]], voxels[ids[1]]


@nb.njit(cache=True)
def umbrella_curv(voxels: nb.float64[:, :], vox_id: nb.int64) -> nb.float64:
    """Computes the umbrella curvature as in equation 9 of "Umbrella Curvature:
    A New Curvature Estimation Method for Point Clouds" by A.Foorginejad and
    K.Khalili
    (https://www.sciencedirect.com/science/article/pii/S2212017313006828)

    Parameters
    ----------
    voxels : np.ndarray
        (N, 3) Voxel coordinates
    vox_id : int
        Index of the voxel w.r.t. which to compute the curvature

    Returns
    -------
    float
        Value of the umbrella curvature at `vox_id`
    """
    # Find the mean direction from that point
    refvox = voxels[vox_id]
    diffs = voxels - refvox
    axis = sm.mean(voxels - refvox, axis=0)
    axis /= np.linalg.norm(axis)

    # Compute the dot product of every displacement vector w.r.t. the axis
    dots = np.zeros(len(diffs), dtype=diffs.dtype)
    for i, diff in enumerate(diffs):
        if i != vox_id:
            dots[i] = np.dot(diff / np.linalg.norm(diff), axis)

    # Find the umbrella curvature (mean angle from the mean direction)
    return abs(np.mean(dots))



"""Module to convert a point cloud into an convex hull envelope."""

import numpy as np
import plotly.graph_objs as go

__all__ = ["hull_trace"]


def hull_trace(
    points,
    color=None,
    intensity=None,
    hovertext=None,
    showscale=False,
    alphahull=0,
    **kwargs,
):
    """Converts a cloud of points into a 3D convex hull.

    This function represents a point cloud by forming a mesh with the points
    that belong to the convex hull of the point cloud.

    Parameters
    ----------
    points : np.ndarray
        (N, 3) Array of point coordinates
    color : Union[str, float, np.ndarray], optional
        Color of hull
    intensity : Union[int, float], optional
        Color intensity of the box along the colorscale axis
    hovertext : Union[int, str, np.ndarray], optional
        Text associated with the cone
    showscale : bool, default False
        If True, show the colorscale of the :class:`plotly.graph_objs.Mesh3d`
    alphahull : float, default 0
        Parameter that sets how to define the hull. 0 is the convex hull,
        larger numbers correspond to alpha-shapes.
    **kwargs : dict, optional
        Additional parameters to pass to the underlying
        :class:`plotly.graph_objs.Mesh3d` object
    """
    # Convert the color provided to a set of intensities, if needed
    if color is not None and not isinstance(color, str):
        assert intensity is None, "Must not provide both `color` and `intensity`."
        intensity = np.full(len(points), color)
        color = None

    # Update hovertemplate style
    hovertemplate = "x: %{x}<br>y: %{y}<br>z: %{z}"
    if hovertext is not None:
        if not np.isscalar(hovertext):
            hovertemplate += "<br>%{text}"
        else:
            hovertemplate += f"<br>{hovertext}"
            hovertext = None

    # Append Mesh3d object
    return go.Mesh3d(
        x=points[:, 0],
        y=points[:, 1],
        z=points[:, 2],
        intensity=intensity,
        alphahull=alphahull,
        showscale=showscale,
        hovertemplate=hovertemplate,
        **kwargs,
    )



def model_dict():
    """Returns dictionary of model classes using name keys (strings).

    Returns
    -------
    dict
        Dictionary of available models
    """

    # from . import spice
    from . import full_chain, graph_spice, grappa, image, uresnet, uresnet_ppn

    # from . import singlep
    # from . import bayes_uresnet
    # from . import vertex
    # Map configuration keys to model/loss tuples
    models = {
        # Full reconstruction chain
        "full_chain": (full_chain.FullChain, full_chain.FullChainLoss),
        # UResNet
        "uresnet": (uresnet.UResNetSegmentation, uresnet.SegmentationLoss),
        # UResNet + PPN
        "uresnet_ppn": (uresnet_ppn.UResNetPPN, uresnet_ppn.UResNetPPNLoss),
        # SPICE
        # "spice": (spice.SPICE, spice.SPICELoss),
        # Graph SPICE
        "graph_spice": (graph_spice.GraphSPICE, graph_spice.GraphSPICELoss),
        # Graph neural network Particle Aggregation (GrapPA)
        "grappa": (grappa.GrapPA, grappa.GrapPALoss),
        # Single Particle Classifier
        "image_class": (image.ImageClassifier, image.ImageClassLoss),
        # Multi Particle Classifier
        # "multip": (singlep.MultiParticleImageClassifier,
        #        singlep.MultiParticleTypeLoss),
        # Bayesian Classifier
        # "bayes_singlep": (singlep.BayesianParticleClassifier,
        #        singlep.ImageClassLoss),
        # Bayesian UResNet
        # "bayesian_uresnet": (bayes_uresnet.BayesianUResNet,
        #        bayes_uresnet.SegmentationLoss),
        # DUQ UResNet
        # "duq_uresnet": (bayes_uresnet.DUQUResNet,
        #        bayes_uresnet.DUQSegmentationLoss),
        # Evidential Classifier
        #'evidential_singlep': (singlep.EvidentialParticleClassifier,
        #        singlep.EvidentialLearningLoss),
        # Evidential Classifier with Dropout
        #'evidential_dropout_singlep': (singlep.BayesianParticleClassifier,
        #        singlep.EvidentialLearningLoss),
        # Deep Single Pass Uncertainty Quantification
        #'duq_singlep': (singlep.DUQParticleClassifier,
        #        singlep.MultiLabelCrossEntropy),
        # Vertex PPN
        #'vertex_ppn': (vertex.VertexPPNChain, vertex.UResNetVertexLoss),
        # Vertex Pointnet
        #'vertex_pointnet': (vertex.VertexPointNet, vertex.VertexPointNetLoss),
    }
    return models


def model_factory(name):
    """
    Returns an instance of a model class based on its name key (string).

    Parameters
    ----------
    name: str
        Key for the model. See source code for list of available models.

    Returns
    -------
    object
    """
    models = model_dict()
    if name not in models:
        raise ValueError("Unknown model name provided: %s" % name)

    return models[name]



import numpy as np


def eval_func_dict():
    funcs = {"mc_dropout": mc_dropout, "default": default}
    return funcs


def construct_eval_func(name):
    funcs = eval_func_dict()
    if name not in funcs:
        raise Exception("Unknown evaluation mode name provided: %s" % name)
    return funcs[name]


def default(net):
    net.eval()


def mc_dropout(net):
    net.eval()
    for m in net.modules():
        if m.__class__.__name__ == "Dropout":
            m.train()


class UniformDistribution:

    def __init__(self, lower_bound, upper_bound, log_scale=False):
        self.lb = lower_bound
        self.ub = upper_bound
        self.log_scale = log_scale

    def sample(self):

        if self.log_scale:
            lb_log = np.log10(self.lb)
            ub_log = np.log10(self.ub)
            exponent = np.random.uniform(lb_log, ub_log)
            return np.power(10, exponent)
        else:
            return np.random.uniform(self.lb, self.ub)



from collections import defaultdict

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch_scatter import scatter_mean

from spine.model.layer.factories import loss_fn_factory, metric_fn_factory
from spine.utils.weighting import get_class_weights

from .misc import *

__all__ = ["NodeEdgeLoss", "EdgeLoss"]


class GraphSPICEEmbeddingLoss(nn.Module):
    """
    Loss function for Sparse Spatial Embeddings Model, with fixed
    centroids and symmetric gaussian kernels.
    """

    name = "embedding"

    def __init__(self, cfg, name="graph_spice_loss"):
        super(GraphSPICEEmbeddingLoss, self).__init__()
        self.loss_config = cfg  # [name]
        self.batch_column = self.loss_config.get("batch_column", 0)

        self.ft_interloss = self.loss_config.get("ft_interloss_margin", 1.5)
        self.sp_interloss = self.loss_config.get("sp_interloss_margin", 0.2)

        self.ft_intraloss = self.loss_config.get("ft_intraloss_margin", 1.0)
        self.sp_intraloss = self.loss_config.get("sp_intraloss_margin", 0.0)

        self.eps = self.loss_config.get("eps", 0.001)

        self.ft_loss_params = self.loss_config.get(
            "ft_loss_params", dict(inter=1.0, intra=1.0, reg=0.1)
        )
        self.sp_loss_params = self.loss_config.get(
            "sp_loss_params", dict(inter=1.0, intra=1.0)
        )

        self.kernel_lossfn_name = self.loss_config.get("kernel_lossfn", "BCE")
        if self.kernel_lossfn_name == "BCE":
            self.kernel_lossfn = nn.BCEWithLogitsLoss(reduction="mean")
        elif self.kernel_lossfn_name == "lovasz_hinge":
            self.kernel_lossfn = LovaszHingeLoss(reduction="none")
        else:
            self.kernel_lossfn = nn.BCEWithLogitsLoss(reduction="none")

        self.seg_lossfn_name = self.loss_config.get("seg_lossfn", "CE")
        if self.seg_lossfn_name == "CE":
            self.seg_loss_fn = nn.CrossEntropyLoss(reduction="mean")
        elif self.seg_lossfn_name == "lovasz_softmax":
            self.seg_loss_fn = LovaszSoftmaxWithLogitsLoss(reduction="mean")
        else:
            self.seg_loss_fn = nn.CrossEntropyLoss(reduction="mean")

    def feature_embedding_loss(self, ft_emb, groups, ft_centroids):
        """
        Compute discriminative feature embedding loss.

        INPUTS:
            - ft_emb (N x F)
            - groups (N)
            - ft_centroids (N_c X F)
        """
        intercluster_loss = inter_cluster_loss(ft_centroids, margin=self.ft_interloss)
        intracluster_loss = intra_cluster_loss(
            ft_emb, ft_centroids, groups, margin=self.ft_intraloss
        )
        reg_loss = torch.mean(torch.norm(ft_centroids, dim=1))
        out = {}
        out["intercluster_loss"] = float(intercluster_loss)
        out["intracluster_loss"] = float(intracluster_loss)
        out["regularization_loss"] = float(reg_loss)
        out["loss"] = (
            self.ft_loss_params["inter"] * intercluster_loss
            + self.ft_loss_params["intra"] * intracluster_loss
            + self.ft_loss_params["reg"] * reg_loss
        )
        return out

    def spatial_embedding_loss(self, sp_emb, groups, sp_centroids):
        """
        Compute spatial centroid regression loss.

        INPUTS:
            - sp_emb (N x D)
            - groups (N)
            - ft_centroids (N_c X F)
        """
        out = {}
        intercluster_loss = inter_cluster_loss(sp_centroids, margin=self.sp_interloss)
        intracluster_loss = intra_cluster_loss(
            sp_emb, sp_centroids, groups, margin=self.sp_intraloss
        )
        out["intercluster_loss"] = float(intercluster_loss)
        out["intracluster_loss"] = float(intracluster_loss)
        out["loss"] = (
            self.sp_loss_params["inter"] * intercluster_loss
            + self.sp_loss_params["intra"] * intracluster_loss
        )

        return out

    def covariance_loss(
        self, sp_emb, ft_emb, cov, groups, sp_centroids, ft_centroids, eps=0.001
    ):

        logits, acc, targets = get_graphspice_logits(
            sp_emb, ft_emb, cov, groups, sp_centroids, ft_centroids, eps
        )
        # Compute kernel score loss
        cov_loss = self.kernel_lossfn(logits, targets)
        return cov_loss, acc

    def occupancy_loss(self, occ, groups):
        """
        INPUTS:
            - occ (N x 1)
            - groups (N)
        """
        bincounts = torch.bincount(groups).float()
        bincounts[bincounts == 0] = 1
        occ_truth = torch.log(bincounts)
        occ_loss = torch.abs(
            torch.gather(occ - occ_truth[None, :], 1, groups.view(-1, 1))
        )
        if len(occ_loss.squeeze().size()) and len(groups.size()):
            occ_loss = scatter_mean(occ_loss.squeeze(), groups)
            # occ_loss = occ_loss[occ_loss > 0]
            return occ_loss.mean()
        else:
            # print(occ_loss.squeeze().size(), groups.size())
            return 0.0

    def combine_multiclass(
        self, sp_embeddings, ft_embeddings, covariance, occupancy, slabels, clabels
    ):
        """
        Wrapper function for combining different components of the loss,
        in particular when clustering must be done PER SEMANTIC CLASS.

        NOTE: When there are multiple semantic classes, we compute the DLoss
        by first masking out by each semantic segmentation (ground-truth/prediction)
        and then compute the clustering loss over each masked point cloud.

        INPUTS:
            features (torch.Tensor): pixel embeddings
            slabels (torch.Tensor): semantic labels
            clabels (torch.Tensor): group/instance/cluster labels

        OUTPUT:
            loss_segs (list): list of computed loss values for each semantic class.
            loss[i] = computed DLoss for semantic class <i>.
            acc_segs (list): list of computed clustering accuracy for each semantic class.
        """
        loss = defaultdict(list)
        loss["loss"] = []
        loss["ft_intra_loss"] = []
        loss["ft_inter_loss"] = []
        loss["ft_reg_loss"] = []
        loss["sp_intra_loss"] = []
        loss["sp_inter_loss"] = []
        loss["cov_loss"] = []
        loss["occ_loss"] = []

        accuracy = defaultdict(float)
        accuracy["accuracy"] = 0.0
        for i in range(5):
            accuracy[f"accuracy_{i}"] = 0.0

        semantic_classes = slabels.unique()
        counts = 0
        for sc in semantic_classes:
            if int(sc) == 4:
                continue
            index = slabels == sc
            sp_emb = sp_embeddings[index]
            ft_emb = ft_embeddings[index]
            cov = covariance[index]
            occ = occupancy[index]
            groups = clabels[index]

            # Remove groups labeled -1
            mask = groups != -1
            sp_emb = sp_emb[mask]
            ft_emb = ft_emb[mask]
            cov = cov[mask]
            occ = occ[mask]
            groups = groups[mask]

            groups_unique, _ = unique_label_torch(groups)
            # print(torch.unique(groups_unique, return_counts=True))
            if groups_unique.shape[0] < 2:
                continue
            sp_centroids = find_cluster_means(sp_emb, groups_unique)
            ft_centroids = find_cluster_means(ft_emb, groups_unique)
            # Get different loss components
            ft_out = self.feature_embedding_loss(ft_emb, groups_unique, ft_centroids)
            sp_out = self.spatial_embedding_loss(sp_emb, groups_unique, sp_centroids)
            cov_loss, acc = self.covariance_loss(
                sp_emb,
                ft_emb,
                cov,
                groups_unique,
                sp_centroids,
                ft_centroids,
                eps=self.eps,
            )
            occ_loss = self.occupancy_loss(occ, groups_unique)
            # TODO: Combine loss with weighting, keep track for logging
            loss["ft_intra_loss"].append(ft_out["intracluster_loss"])
            loss["ft_inter_loss"].append(ft_out["intercluster_loss"])
            loss["ft_reg_loss"].append(ft_out["regularization_loss"])
            loss["sp_intra_loss"].append(sp_out["intracluster_loss"])
            loss["sp_inter_loss"].append(sp_out["intercluster_loss"])
            loss["cov_loss"].append(float(cov_loss))
            loss["occ_loss"].append(float(occ_loss))
            loss["loss"].append(ft_out["loss"] + sp_out["loss"] + cov_loss + occ_loss)
            # TODO: Implement train-time accuracy estimation
            accuracy["accuracy_{}".format(int(sc))] = acc
            accuracy["accuracy"] += acc
            counts += 1

        if counts > 0:
            accuracy["accuracy"] /= counts
            for i in range(5):
                accuracy[f"accuracy_{i}"] /= counts

        return loss, accuracy

    def forward(self, out, segment_label, cluster_label):

        num_gpus = len(segment_label)
        loss = defaultdict(list)
        accuracy = defaultdict(list)

        segmentationLayer = "segmentation" in out
        for i in range(num_gpus):
            slabels = segment_label[i][:, -1]
            # coords = segment_label[i][:, :3].float()
            # if torch.cuda.is_available():
            #    coords = coords.cuda()
            slabels = slabels.long()
            clabels = cluster_label[i][:, -1].long()
            # print(clabels)
            batch_idx = segment_label[i][:, self.batch_column]
            sp_embedding = out["spatial_embeddings"][i]
            ft_embedding = out["feature_embeddings"][i]
            covariance = out["covariance"][i]
            occupancy = out["occupancy"][i]
            if segmentationLayer:
                segmentation = out["segmentation"][i]
            # nbatch = batch_idx.unique().shape[0]

            for bidx in batch_idx.unique(sorted=True):
                batch_mask = batch_idx == bidx
                sp_embedding_batch = sp_embedding[batch_mask]
                ft_embedding_batch = ft_embedding[batch_mask]
                if segmentationLayer:
                    segmentation_batch = segmentation[batch_mask]
                slabels_batch = slabels[batch_mask]
                clabels_batch = clabels[batch_mask]
                covariance_batch = covariance[batch_mask]
                occupancy_batch = occupancy[batch_mask]

                if segmentationLayer:
                    loss_seg = self.seg_loss_fn(segmentation_batch, slabels_batch)
                    acc_seg = float(
                        torch.sum(
                            torch.argmax(segmentation_batch, dim=1) == slabels_batch
                        )
                    ) / float(segmentation_batch.shape[0])

                loss_class, acc_class = self.combine_multiclass(
                    sp_embedding_batch,
                    ft_embedding_batch,
                    covariance_batch,
                    occupancy_batch,
                    slabels_batch,
                    clabels_batch,
                )
                for key, val in loss_class.items():
                    loss[key].append(sum(val) / len(val) if len(val) else 0.0)
                for s, acc in acc_class.items():
                    accuracy[s].append(acc)

                if segmentationLayer:
                    loss["gs_loss_seg"].append(loss_seg)
                    accuracy["gs_acc_seg"].append(acc_seg)

        loss_avg = {}
        acc_avg = defaultdict(float)

        for key, val in loss.items():
            if sum(val) > 0:
                loss_avg[key] = sum(val) / len(val)
            else:
                loss_avg[key] = 0.0
        if segmentationLayer:
            loss_avg["loss"] += loss_avg["gs_loss_seg"]
        for key, val in accuracy.items():
            if sum(val) > 0:
                acc_avg[key] = sum(val) / len(val)
            else:
                acc_avg[key] = 1.0

        res = {}
        res.update(loss_avg)
        res.update(acc_avg)

        return res


class NodeEdgeLoss(torch.nn.modules.loss._Loss):
    """
    Combined Node + Edge Loss
    """

    name = "node_edge"

    def __init__(
        self,
        invert=True,
        edge_loss=None,
        use_cluster_labels=True,
        embedding_loss_weight=1.0,
        **kwargs,
    ):
        super(NodeEdgeHybridLoss, self).__init__()

        self.loss_fn = GraphSPICEEmbeddingLoss(kwargs)
        if edge_loss is None:
            edge_loss = {}
        self.edge_loss = edge_loss
        self.invert = invert
        self.edge_loss = WeightedEdgeLoss(invert=self.invert, **self.edge_loss)

        self.metric_fn = IoUScore()
        self.use_cluster_labels = use_cluster_labels
        self.embedding_loss_weight = embedding_loss_weight

    def forward(self, result, segment_label, cluster_label):

        group_label = [cluster_label[0][:, [0, 1, 2, 3, 5]]]

        res = self.loss_fn(result, segment_label, group_label)
        edge_score = result["gs_edge_attr"][0].squeeze()
        x = edge_score
        pred = x >= 0

        iou, edge_loss = 0, 0

        if self.use_cluster_labels:
            edge_truth = result["gs_edge_label"][0].squeeze()
            # print(edge_score.squeeze(), edge_truth, edge_score.shape, edge_truth.shape)
            edge_loss = self.edge_loss(edge_score, edge_truth.float())
            edge_loss = edge_loss.mean()

            if self.invert:
                edge_truth = torch.logical_not(edge_truth.long())

            iou = self.metric_fn(pred, edge_truth)
        # iou2 = self.metric_fn(~pred, ~edge_truth)

        res["edge_accuracy"] = iou
        if "loss" in res:
            res["loss"] = res["loss"] * self.embedding_loss_weight + edge_loss
        else:
            res["loss"] = edge_loss
        res["edge_loss"] = float(edge_loss)

        return res


class EdgeLoss(torch.nn.modules.loss._Loss):
    """Loss applied to edge scores produced by a GraphSPICE.

    This loss simply treats the edge score as logit predictions and compares
    them to edge labels (simple classification task).
    """

    name = "edge"

    def __init__(
        self,
        loss="log_dice",
        invert=True,
        balance_loss=False,
        equal_sampling=False,
        min_sample_edges=1000,
        metric="iou",
    ):
        """Initialize the loss function.

        Parameters
        ----------
        loss : str, default 'log_dice'
            Loss functional used to train edge scores
        invert : bool, default True
            If `True`, on edges are labeled as 0 and off edges as 1
        balance_loss : bool, default True
            Whether to weight the loss to account for class imbalance
        equal_sampling : bool, default False
            If `True`, sample the same number of edges from each label class
        min_sample_edges : int, default 1000
            If sampling evenly, minimum number of edges to sample with replacement
        """
        # Initialize the parent class
        super().__init__()

        # Store parameters
        self.invert = invert
        self.balance_loss = balance_loss
        self.equal_sampling = equal_sampling
        self.min_sample_edges = min_sample_edges

        # Intialize the loss function
        self.loss_fn = loss_fn_factory(loss, reduction="none")

        # Intitliaze the extra metric, if required
        self.metric_fn = None
        if metric is not None:
            self.metric_fn = metric_fn_factory(metric)

    def sample_edges(self, edge_score, edge_truth):
        """Subsample a set number of edges from each edge label class.

        Parameters
        ----------
        edge_scores : torch.Tensor
            Edge score predictions
        edge_label : torch.Tensor
            Edge labels

        Returns
        -------
        edge_scores : torch.Tensor
            Subsampled edge predictions
        edge_label : torch.Tensor
            Subsampled edge labels
        """
        with torch.no_grad():
            num_gaps, num_bridges = torch.bincount(edge_truth)
            n = max(num_gaps, self.min_sampled_edges)
            perm = torch.randperm(n).cuda()
            index = torch.arange(edge_truth.shape[0], dtype=torch.long).cuda()

        if num_gaps == 0:
            mask = index[perm]
            sampled_score, sampled_truth = edge_score[mask], edge_truth[mask]
        else:
            gap_indices = index[edge_truth == 0]
            brg_indices = index[edge_truth == 1]

            sampled_brgs = brg_indices[perm]
            index = torch.cat([gap_indices, brg_indices])

            sampled_score, sampled_truth = edge_score[index], edge_truth[index]

        return sampled_score, sampled_truth

    def forward(self, clust_label, edge_attr, edge_label, **kwargs):
        """Applies the edge loss to a batch of data.

        Parameters
        ----------
        clust_label : TensorBatch
            (N, 1 + D + N_c) Tensor of cluster labels
            - N_c is is the number of cluster labels
        edge_attr : TensorBatch
            (E) Edge scores
        edge_label : TensorBatch
            (E) Edge binary labels
        **kwargs : dict, optional
            Additional upstream model outputs not used in this loss

        Returns
        -------
        loss : torch.Tensor
            Value of the loss
        accuracy : float
            Value of the edge-wise classification accuracy
        iou : float
            IoU accuracy metric
        count : int
            Number of edges the loss was applied to
        """
        # Extract the raw tensors from the batches
        edge_attr = edge_attr.tensor.flatten()
        edge_pred = (edge_attr > 0.0).long()
        edge_label = edge_label.tensor

        # If requested, extract an equal number of samples from scores/labels
        if self.equal_sampling:
            edge_attr, edge_label = self.sample_edges(edge_attr, edge_label)

        # If requested, make off connections the positive label
        if self.invert:
            edge_label = torch.logical_not(edge_label).long()

        # Apply the loss function
        loss = self.loss_fn(edge_attr, edge_label)

        # If requested, apply the weights
        if self.balance_loss:
            weight = get_class_weights(edge_label, 2, per_class=False)
            loss *= weight

        loss = loss.mean()

        # Evaluate the accuracy
        num_edges = len(edge_pred)
        accuracy = 1.0
        if num_edges > 0:
            accuracy = (edge_pred == edge_label).sum() / num_edges

        metric = {}
        if self.metric_fn is not None:
            metric = {self.metric_fn.name: self.metric_fn(edge_pred, edge_label)}

        # Prepare and return the result dictionary
        return {"loss": loss, "accuracy": accuracy, **metric}



"""CUDA device management and distributed training setup.

This module provides utilities for configuring GPU devices and setting up
distributed training environments with graceful PyTorch unavailability handling.
"""

import os

from ..conditional import TORCH_AVAILABLE, torch

__all__ = ["set_visible_devices"]


def set_visible_devices(gpus=None, world_size=None, **kwargs):
    """Sets the number of visible CUDA devices based on the base configuration.

    Parameters
    ----------
    gpus : List[int], optional
        List of indexes of GPUs to expose to the model
    world_size : int, optional
        Number of GPUs to use in the model
    **kwargs : dict, optional
        Additional arguments passed to the base configuration

    Returns
    -------
    int
        World size
    """
    # Check if torch is available for GPU operations
    if not TORCH_AVAILABLE:
        if gpus is not None or world_size is not None and world_size > 0:
            raise ImportError(
                "PyTorch is required for GPU operations. "
                "Install with: pip install spine-ml[model]"
            )
        return 0  # CPU only

    # Fetch the list of devices to expose, set the world size accordingly
    if world_size is None and gpus is None:
        # If no GPU request is put in, run on CPU
        world_size = 0
    else:
        # Otherwise, harmonize
        assert (world_size is None) or (gpus is None) or len(gpus) == world_size, (
            f"The world size ({world_size}) does not match the "
            f"number of exposed GPUs ({len(gpus)})."
        )
        world_size = world_size if world_size is not None else len(gpus)
        gpus = gpus if gpus is not None else list(range(world_size))

    # Set the visible CUDA devices
    if not os.environ.get("CUDA_VISIBLE_DEVICES", None) and gpus is not None:
        # If it is not yet set, do it
        os.environ["CUDA_VISIBLE_DEVICES"] = ",".join([str(g) for g in gpus])

    # Make sure the world size is consistent with the number of visible GPUs
    if world_size > 0:
        assert (
            torch.cuda.is_available
        ), "Cannot use distributed training without access to GPUs."

        visible_devices = torch.cuda.device_count()
        assert world_size <= visible_devices, (
            f"The number of GPUs requested ({world_size}) exceeds the "
            f"number of visible devices ({visible_devices})."
        )

    # Return the world size
    return world_size



"""API version and metadata constants for SPINE configuration loading.

This module defines the configuration API version and metadata schema.
"""

# Current API version
API_VERSION = "1.0"

# Metadata schema
META_KEY = "__meta__"

# Metadata fields
META_VERSION = "version"
META_DATE = "date"
META_DESCRIPTION = "description"
META_COMPATIBLE_WITH = "compatible_with"
META_TAGS = "tags"
META_KIND = "kind"  # "bundle" or "mod"
META_STRICT = "strict"  # "warn" or "error"
META_LIST_APPEND = "list_append"  # "append" or "unique"

# Default values
DEFAULT_KIND = "bundle"
DEFAULT_STRICT = "warn"  # warn for mods, error for bundles (see loader)
DEFAULT_LIST_APPEND = "append"  # current behavior

# Valid values
VALID_KINDS = {"bundle", "mod"}
VALID_STRICT_MODES = {"warn", "error"}
VALID_LIST_APPEND_MODES = {"append", "unique"}



# Copyright (c) Facebook, Inc. and its affiliates.
# Modified by Bowen Cheng from https://github.com/facebookresearch/detr/blob/master/models/detr.py
# Modified for lartpc_spine3d

import torch
import torch.nn.functional as F
from scipy.optimize import linear_sum_assignment
from torch import nn

from spine.model.layer.cluster.losses.misc import LovaszHingeLoss, iou_batch
from spine.utils.globals import *


class LinearSumAssignmentLoss(nn.Module):

    def __init__(self, weight_dice=2.0, weight_ce=5.0, mode="dice"):
        super(LinearSumAssignmentLoss, self).__init__()
        self.weight_dice = weight_dice
        self.weight_ce = weight_ce

        self.lovasz = LovaszHingeLoss()
        self.mode = mode
        print(f"Setting LinearSumAssignment loss to '{self.mode}'")

    def compute_accuracy(self, masks, targets, indices):
        with torch.no_grad():
            valid_masks = masks[:, indices[0]] > 0
            valid_targets = targets[:, indices[1]] > 0.5
            iou = iou_batch(valid_masks, valid_targets, eps=1e-6)
            return float(iou)

    def forward(self, masks, targets):

        with torch.no_grad():
            dice_loss = batch_dice_loss(masks.T, targets.T)
            ce_loss = batch_sigmoid_ce_loss(masks.T, targets.T)
            cost_matrix = self.weight_dice * dice_loss + self.weight_ce * ce_loss
            indices = linear_sum_assignment(cost_matrix.detach().cpu())

        if self.mode == "log_dice":
            dice_loss = log_dice_loss_flat(masks[:, indices[0]], targets[:, indices[1]])
        elif self.mode == "dice":
            dice_loss = dice_loss_flat(masks[:, indices[0]], targets[:, indices[1]])
        # elif self.mode == 'lovasz':
        #     dice_loss = self.lovasz(masks[:, indices[0]], targets[:, indices[1]])
        else:
            raise ValueError(f"LSA loss mode {self.mode} is not supported!")
        ce_loss = sigmoid_ce_loss(masks.T[indices[0]], targets.T[indices[1]])
        loss = self.weight_dice * dice_loss + self.weight_ce * ce_loss
        acc = self.compute_accuracy(masks, targets, indices)

        return loss, acc


class CEDiceLoss(nn.Module):

    def __init__(self, weight_dice=1.0, weight_ce=1.0, mode="dice"):
        super(CEDiceLoss, self).__init__()
        self.weight_dice = weight_dice
        self.weight_ce = weight_ce
        self.mode = mode
        print(f"Setting LinearSumAssignment loss to '{self.mode}'")

    def compute_accuracy(self, masks, targets):
        with torch.no_grad():
            valid_masks = masks > 0
            valid_targets = targets > 0.5

            print(masks.sum(dim=0))
            print(targets.sum(dim=0))

            iou = iou_batch(valid_masks, valid_targets, eps=1e-6)
            return float(iou)

    def forward(self, masks, targets):

        dice_loss = dice_loss_flat(masks, targets)
        # if self.mode == 'log_dice':
        #     dice_loss = batch_log_dice_loss(masks.T[indices[0]], targets.T[indices[1]])
        # elif self.mode == 'dice':
        #     dice_loss = batch_dice_loss(masks.T[indices[0]], targets.T[indices[1]])
        # elif self.mode == 'lovasz':
        #     dice_loss = self.lovasz(masks[:, indices[0]], targets[:, indices[1]])
        # else:
        #     raise ValueError(f"LSA loss mode {self.mode} is not supported!")
        ce_loss = sigmoid_ce_loss(masks.T, targets.T)
        loss = self.weight_dice * dice_loss + self.weight_ce * ce_loss
        acc = self.compute_accuracy(masks, targets)

        return loss, acc


@torch.jit.script
def get_instance_masks(cluster_label: torch.LongTensor, max_num_instances: int = -1):
    """Given integer coded cluster instance labels, construct a
    (N x max_num_instances) bool tensor in which each colume is a
    binary instance mask.

    """
    groups, counts = torch.unique(cluster_label, return_counts=True)
    if max_num_instances < 0:
        max_num_instances = groups.shape[0]
    instance_masks = torch.zeros((cluster_label.shape[0], max_num_instances)).to(
        device=cluster_label.device, dtype=torch.bool
    )
    perm = torch.argsort(counts, descending=True)[:max_num_instances]

    for i, group_id in enumerate(groups[perm]):
        instance_masks[:, i] = (cluster_label == group_id).to(torch.bool)

    return instance_masks


@torch.jit.script
def get_instance_masks_from_queries(
    cluster_label: torch.LongTensor, query_index: torch.Tensor
):
    max_num_instances = query_index.shape[0]
    instance_masks = torch.zeros((cluster_label.shape[0], max_num_instances)).to(
        device=cluster_label.device, dtype=torch.bool
    )
    for i, qidx in enumerate(query_index):
        instance_masks[:, i] = (cluster_label == cluster_label[qidx]).to(torch.bool)

    return instance_masks


def dice_loss(logits, targets):
    """

    Parameters
    ----------
    logits: (N x num_queries)
    targets: (N x num_queries)
    """
    num_masks = logits.shape[1]
    scores = torch.sigmoid(logits)
    numerator = (2 * scores * targets).sum(dim=0)
    denominator = scores.sum(dim=0) + targets.sum(dim=0)
    return (1 - (numerator + 1) / (denominator + 1)).sum() / num_masks


@torch.jit.script
def batch_dice_loss(inputs: torch.Tensor, targets: torch.Tensor):
    """
    Compute the DICE loss, similar to generalized IOU for masks
    Args:
        inputs: (num_masks, num_points) Tensor
        targets: (num_masks, num_points) Tensor
    """
    scores = inputs.sigmoid()
    scores = scores.flatten(1)
    numerator = 2 * torch.einsum("nc,mc->nm", scores, targets)
    denominator = scores.sum(-1)[:, None] + targets.sum(-1)[None, :]
    loss = 1 - (numerator + 1) / (denominator + 1)
    return loss


@torch.jit.script
def batch_log_dice_loss(inputs: torch.Tensor, targets: torch.Tensor):
    """
    Compute the DICE loss, similar to generalized IOU for masks
    Args:
        inputs: (num_masks, num_points) Tensor
        targets: (num_masks, num_points) Tensor
    """
    scores = inputs.sigmoid()
    scores = scores.flatten(1)
    numerator = 2 * torch.einsum("nc,mc->nm", scores, targets)
    denominator = scores.sum(-1)[:, None] + targets.sum(-1)[None, :]
    loss = -torch.log((numerator + 1) / (denominator + 1))
    return loss


@torch.jit.script
def batch_sigmoid_ce_loss(inputs: torch.Tensor, targets: torch.Tensor):
    """
    Args:
        inputs: A float tensor of arbitrary shape.
                The predictions for each example.
        targets: A float tensor with the same shape as inputs. Stores the binary
                 classification label for each element in inputs
                (0 for the negative class and 1 for the positive class).
    Returns:
        Loss tensor
    """
    hw = inputs.shape[1]

    pos = F.binary_cross_entropy_with_logits(
        inputs, torch.ones_like(inputs), reduction="none"
    )
    neg = F.binary_cross_entropy_with_logits(
        inputs, torch.zeros_like(inputs), reduction="none"
    )

    loss = torch.einsum("nc,mc->nm", pos, targets) + torch.einsum(
        "nc,mc->nm", neg, (1 - targets)
    )

    return loss / hw


@torch.jit.script
def sigmoid_ce_loss(inputs: torch.Tensor, targets: torch.Tensor):
    """
    Args:
        inputs: A float tensor of arbitrary shape.
                The predictions for each example.
        targets: A float tensor with the same shape as inputs. Stores the binary
                 classification label for each element in inputs
                (0 for the negative class and 1 for the positive class).
    Returns:
        Loss tensor
    """
    num_masks = inputs.shape[0]
    loss = F.binary_cross_entropy_with_logits(inputs, targets, reduction="none")
    return loss.mean(1).sum() / num_masks


@torch.jit.script
def dice_loss_flat(logits, targets):
    """

    Parameters
    ----------
    logits: (N x num_queries)
    targets: (N x num_queries)
    """
    num_masks = logits.shape[1]
    scores = torch.sigmoid(logits)
    numerator = (2 * scores * targets).sum(dim=0)
    denominator = scores.sum(dim=0) + targets.sum(dim=0)
    return (1 - (numerator + 1) / (denominator + 1)).sum() / num_masks


@torch.jit.script
def log_dice_loss_flat(logits, targets):
    """

    Parameters
    ----------
    logits: (N x num_queries)
    targets: (N x num_queries)
    """
    num_masks = logits.shape[1]
    scores = torch.sigmoid(logits)
    numerator = (2 * scores * targets).sum(dim=0)
    denominator = scores.sum(dim=0) + targets.sum(dim=0)
    return (-torch.log(1 - (numerator + 1) / (denominator + 1))).sum() / num_masks



"""Module which does connected-components (dense) clustering using DBSCAN."""

import numpy as np
import torch

from spine.data import IndexBatch, TensorBatch
from spine.math.cluster import DBSCAN as spine_dbscan
from spine.utils.globals import (
    COORD_COLS,
    COORD_END_COLS,
    COORD_START_COLS,
    DELTA_SHP,
    MICHL_SHP,
    PPN_SHAPE_COL,
    SHOWR_SHP,
    TRACK_SHP,
)
from spine.utils.point_break_clustering import PointBreakClusterer
from spine.utils.ppn import PPNPredictor


class DBSCAN(torch.nn.Module):
    """Uses DBSCAN to find locally-dense particle fragments.

    It uses SPINE's numba-accelerated DBSCAN implementation to fragment each of
    the particle shapes into dense instances. Runs DBSCAN on each requested
    semantic class separately, in one of three ways:
    - Run pure DBSCAN on all the voxels in that class
    - Runs DBSCAN on PPN point-masked voxels and then associates the
      leftovers based on proximity to existing instances.
    - Use a graph-based method to cluster tracks based on PPN vertices. This
      technique can only be used on tracks.
    """

    def __init__(
        self,
        eps=1.8,
        min_samples=1,
        min_size=3,
        metric="euclidean",
        shapes=[SHOWR_SHP, TRACK_SHP, MICHL_SHP, DELTA_SHP],
        break_shapes=[TRACK_SHP],
        break_mask_radius=5.0,
        break_track_method="masked_dbscan",
        use_label_break_points=False,
        track_include_delta=False,
        ppn_predictor={},
    ):
        """Initialize the DBSCAN clustering algorithm.

        Parameters
        ----------
        eps : float, default 1.8
            The maximum distance between two samples for one to be considered
            as in the neighborhood of the other.
        min_samples : int, default 1
            The number of samples (or total weight) in a neighborhood for a
            point to be considered as a core point.
        min_size : int, default 3
            Minimum cluster size to stored in the final list of DBSCAN clusters
        metric : str, default 'euclidean'
            Metric used to compute the pair-wise distances between space points
        shapes : List[int], default [0, 1, 2, 3]
            List of semantic classes to run DBSCAN on
        break_shapes : List[int], default [1]
            List of semantic shapes for which to use PPN to break down
        break_mask_radius : str, default 5.0
            If using particle points to break up instances further, specifies
            the radius around each particle point which gets masked
        break_track_method : str, default 'masked_dbscan'
            If using particle points to break up tracks, specifies the method
        use_label_break_points : bool, default False
            Whether to use label points to break instances
        track_include_delta : bool, default False
            If `True`, include delta points along with track point when
            running DBSCAN on track points (limits artificial track breaks)
        ppn_predictor : cfg, optional
            PPN post-processing configuration
        """
        # Initialize the parent class
        super().__init__()

        # Store the DBSCAN clustering parameters
        self.eps = eps
        self.min_samples = min_samples
        self.min_size = min_size
        self.metric = metric
        self.shapes = shapes
        self.break_shapes = break_shapes
        self.break_mask_radius = break_mask_radius
        self.break_track_method = break_track_method
        self.track_include_delta = track_include_delta

        # If the constants are provided as scalars, turn them into lists
        assert not np.isscalar(shapes), "Semantic classes should be provided as a list."
        for attr in ["eps", "min_samples", "min_size", "metric", "break_mask_radius"]:
            if np.isscalar(getattr(self, attr)):
                setattr(self, attr, len(shapes) * [getattr(self, attr)])
            else:
                assert len(getattr(self, attr)) == len(shapes), (
                    f"The number of `{attr}` values does not match the "
                    "number shapes to cluster."
                )

        # Instantiate the PPN post-processor, if needed
        self.use_label_break_points = use_label_break_points
        assert not np.isscalar(
            break_shapes
        ), "Semantic classes to break should be provided as a list."
        if len(break_shapes) and not use_label_break_points:
            assert ppn_predictor is not None, (
                "If shapes are to be broken up using PPN points, "
                "must provide a PPN predictor configuration."
            )
            self.ppn_predictor = PPNPredictor(**ppn_predictor)

        # Initialize one clustering algorithm per class
        self.clusterers = []
        for k, c in enumerate(shapes):
            if c not in break_shapes:
                dbscan = spine_dbscan(
                    eps=self.eps[k],
                    min_samples=self.min_samples[k],
                    metric=self.metric[k],
                )
                clusterer = lambda x, _: dbscan.fit_predict(x)
            else:
                method = break_track_method
                if c != TRACK_SHP:
                    method = "masked_dbscan"
                clusterer = PointBreakClusterer(
                    eps=self.eps[k],
                    min_samples=self.min_samples[k],
                    metric=self.metric[k],
                    method=method,
                    mask_radius=self.break_mask_radius[k],
                )

            self.clusterers.append(clusterer)

    def forward(self, data, seg_pred, coord_label=None, **ppn_result):
        """Pass a batch of data through DBSCAN to form space clusters.

        Parameters
        ----------
        data : TensorBatch
            (N, 1 + D + N_f) Tensor of voxel/value pairs
            - N is the the total number of voxels in the image
            - 1 is the batch ID
            - D is the number of dimensions in the input image
            - N_f is 1 (charge/energy) if the clusters (`clusts`) are provided,
              or it needs to contain cluster labels to build them on the fly
        seg_pred : TensorBatch
            (N) Segmentation value for each data point
        coord_label : TensorBatch, optional
            Location of the true particle points
        **ppn_result : dict, optional
            Dictionary of outputs from the PPN model
        """
        # If some shapes must be broken up at their points of interest,
        # fetch them from the relevant location.
        points, point_shapes = None, None
        if len(self.break_shapes):
            if self.use_label_break_points:
                assert coord_label is not None, (
                    "If label points are to be used to break instance, "
                    "must provide them."
                )
                points = torch.cat(
                    (
                        coord_label.tensor[:, COORD_START_COLS],
                        coord_label.tensor[:, COORD_END_COLS],
                    ),
                    dim=1,
                ).reshape(-1, 3)
                point_shapes = torch.repeat_interleave(
                    coord_label.tensor[:, SHAPE_COL], 2
                )
                points = TensorBatch(points, 2 * coord_label.counts)
                point_shapes = TensorBatch(point_shapes, 2 * coord_label.counts)
            else:
                ppn_points = self.ppn_predictor(**ppn_result)
                points = TensorBatch(
                    ppn_points.tensor[:, COORD_COLS], ppn_points.counts
                )
                point_shapes = TensorBatch(
                    ppn_points.tensor[:, PPN_SHAPE_COL], ppn_points.counts
                )

        # Bring everything to numpy (DBSCAN cannot run on tensors)
        data_np = data.to_numpy()
        seg_pred_np = seg_pred.to_numpy()
        if points is not None:
            points_np = points.to_numpy()
            point_shapes_np = point_shapes.to_numpy()

        # Loop over the entries in the batch
        offsets = data.edges[:-1]
        clusts, shapes, counts, single_counts = [], [], [], []
        for b in range(data.batch_size):
            # Fetch the necessary data products, in numpy format
            voxels_b = data_np[b][:, COORD_COLS]
            seg_pred_b = seg_pred_np[b]
            if points is not None:
                points_b = points_np[b]
                point_shapes_b = point_shapes_np[b]

                # Exclude delta points, they do not help with clustering
                points_b = points_b[point_shapes_b != DELTA_SHP]

            # Loop over the shapes to cluster
            clusts_b, counts_b, shapes_b = [], [], []
            for k, s in enumerate(self.shapes):
                # Restrict the voxels to the current class
                break_class = s in self.break_shapes
                shape_mask = seg_pred_b == s
                if s == TRACK_SHP and break_class and self.track_include_delta:
                    shape_mask |= seg_pred_b == DELTA_SHP

                shape_index = np.where(shape_mask)[0]
                if not len(shape_index):
                    continue

                # Run clustering
                voxels_b_s = voxels_b[shape_index]
                labels = self.clusterers[k](voxels_b_s, points_b)

                # If delta points were added to track points, remove them
                if s == TRACK_SHP and break_class and self.track_include_delta:
                    labels[seg_pred_b[shape_index] == DELTA_SHP] = -1

                # Build clusters for this class
                clusts_b_s = []
                for c in np.unique(labels):
                    clust = np.where(labels == c)[0]
                    if c > -1 and len(clust) > self.min_size[k]:
                        clusts_b_s.append(int(offsets[b]) + clust)
                        counts_b.append(len(clust))

                clusts_b.extend(clusts_b_s)
                shapes_b.append(s * np.ones(len(clusts_b_s), dtype=np.int64))

            # Update the output
            clusts.extend(clusts_b)
            shapes.extend(shapes_b)
            counts.append(len(clusts_b))
            single_counts.extend(counts_b)

        # Initialize an IndexBatch and return it
        clusts_nb = np.empty(len(clusts), dtype=object)
        clusts_nb[:] = clusts

        index = IndexBatch(clusts_nb, offsets, counts, single_counts)
        if len(shapes):
            shapes = TensorBatch(np.concatenate(shapes), counts)
        else:
            shapes = TensorBatch(np.empty(0, dtype=np.int64), counts)

        return index, shapes



"""I/O tools for reading and writing neutrino physics data.

This module provides comprehensive data input/output capabilities for various file formats
commonly used in neutrino physics experiments, with support for both PyTorch-dependent
and independent operations.

**Core I/O Functionality:**
- `core`: Framework-independent I/O operations (LArCV, HDF5, ROOT)
- `torch`: PyTorch-specific data loading and batch processing

**File Format Support:**
- **LArCV**: ROOT-based sparse tensor format for liquid argon detectors
- **HDF5**: Hierarchical data format with compression and metadata
- **ROOT**: High-energy physics standard format
- **NumPy**: Native Python array format

**Key Components:**
- **Parsers**: Extract and transform data from various file formats
- **Collaters**: Batch multiple samples together for ML training
- **Samplers**: Control data sampling strategies and ordering
- **Writers**: Output processed data and results

**Writing Custom I/O Functions:**
1. Add a parser in the appropriate parser module
2. Implement or modify collate functions for batching
3. Optionally create custom sampling functions

**Example HDF5 Configuration:**
```yaml
io:
  writer:
    name: HDF5Writer
    file_name: output.h5
    input_keys: ['input_data', 'truth_labels']
    result_keys: ['predictions', 'metrics']
```

**Features:**
- Lazy loading for memory efficiency
- Multi-file dataset handling
- Automatic data type conversions
- Metadata preservation
- Progress tracking and logging
- Error recovery and validation

**Conditional Imports:**
PyTorch-dependent functionality is conditionally imported and available
only when PyTorch is installed. Check `TORCH_IO_AVAILABLE` flag.
"""

# Always import PyTorch-independent core functionality
from .core import *

# Conditionally import PyTorch-dependent functionality
try:
    from .torch import *

    TORCH_IO_AVAILABLE = True
except ImportError:
    TORCH_IO_AVAILABLE = False



import MinkowskiEngine as ME
import MinkowskiFunctional as MF
import numpy as np
import torch
import torch.nn as nn

from .blocks import *
from .configuration import setup_cnn_configuration
from .nonlinearities import MinkowskiLeakyReLU


class FPN(torch.nn.Module):
    """Feature Pyramid Network (FPN).

    FPNs are a different implementation of the concept behind unets.

    Configurations
    --------------
    depth : int
        Depth of FPN, also corresponds to how many times we down/upsample.
    num_filters : int
        Number of filters in the first convolution of FPN.
        Will increase linearly with depth.
    reps : int, optional
        Convolution block repetition factor
    kernel_size : int, optional
        Kernel size for the SC (sparse convolutions for down/upsample).
    input_kernel : int, optional
        Receptive field size for very first convolution after input layer.
    """

    def __init__(self, cfg, name="fpn"):
        super(FPN, self).__init__()
        setup_cnn_configuration(self, cfg, name)

        model_cfg = cfg["modules"][name]

        # UResNet Configurations
        self.reps = model_cfg.get("reps", 2)
        self.depth = model_cfg.get("depth", 5)
        self.num_filters = model_cfg.get("num_filters", 16)
        self.nPlanes = [i * self.num_filters for i in range(1, self.depth + 1)]
        # self.kernel_size = cfg.get('kernel_size', 3)
        # self.downsample = cfg.get(downsample, 2)
        self.input_kernel = model_cfg.get("input_kernel", 3)

        # Initialize Input Layer
        self.input_layer = ME.MinkowskiConvolution(
            in_channels=self.num_input,
            out_channels=self.num_filters,
            kernel_size=self.input_kernel,
            stride=1,
            dimension=self.D,
        )

        # Initialize Encoder
        self.encoding_conv = []
        self.encoding_block = []
        for i, F in enumerate(self.nPlanes):
            m = []
            for _ in range(self.reps):
                m.append(ResNetBlock(F, F, dimension=self.D, leakiness=self.leakiness))
            m = nn.Sequential(*m)
            self.encoding_block.append(m)
            m = []
            if i < self.depth - 1:
                m.append(ME.MinkowskiBatchNorm(F))
                m.append(MinkowskiLeakyReLU(negative_slope=self.leakiness))
                m.append(
                    ME.MinkowskiConvolution(
                        in_channels=self.nPlanes[i],
                        out_channels=self.nPlanes[i + 1],
                        kernel_size=2,
                        stride=2,
                        dimension=self.D,
                    )
                )
            m = nn.Sequential(*m)
            self.encoding_conv.append(m)
        self.encoding_conv = nn.Sequential(*self.encoding_conv)
        self.encoding_block = nn.Sequential(*self.encoding_block)

        # Lateral Connections
        self.lateral = []
        for i, F in enumerate(self.nPlanes[-2::-1]):
            self.lateral.append(ME.MinkowskiLinear(F, F))
        self.lateral = nn.Sequential(*self.lateral)

        # Initialize Decoder
        self.decoding_block = []
        self.decoding_conv = []
        for i in range(self.depth - 2, -1, -1):
            m = []
            m.append(ME.MinkowskiBatchNorm(self.nPlanes[i + 1]))
            m.append(MinkowskiLeakyReLU(negative_slope=self.leakiness))
            m.append(
                ME.MinkowskiConvolutionTranspose(
                    in_channels=self.nPlanes[i + 1],
                    out_channels=self.nPlanes[i],
                    kernel_size=2,
                    stride=2,
                    dimension=self.D,
                )
            )
            m = nn.Sequential(*m)
            self.decoding_conv.append(m)
            m = []
            for j in range(self.reps):
                m.append(
                    ResNetBlock(
                        self.nPlanes[i],
                        self.nPlanes[i],
                        dimension=self.D,
                        leakiness=self.leakiness,
                    )
                )
            m = nn.Sequential(*m)
            self.decoding_block.append(m)
        self.decoding_block = nn.Sequential(*self.decoding_block)
        self.decoding_conv = nn.Sequential(*self.decoding_conv)

        print(
            "Total Number of Trainable Parameters = {}".format(
                sum(p.numel() for p in self.parameters() if p.requires_grad)
            )
        )

    def encoder(self, x):
        """
        Vanilla FPN Encoder.

        INPUTS:
            - x (SparseTensor): MinkowskiEngine SparseTensor

        RETURNS:
            - result (dict): dictionary of encoder output with
            intermediate feature planes:
              1) encoderTensors (list): list of intermediate SparseTensors
              2) finalTensor (SparseTensor): feature tensor at
              deepest layer.
        """
        x = self.input_layer(x)
        encoderTensors = [x]
        for i, layer in enumerate(self.encoding_block):
            x = self.encoding_block[i](x)
            encoderTensors.append(x)
            x = self.encoding_conv[i](x)

        result = {"encoderTensors": encoderTensors, "finalTensor": x}
        return result

    def decoder(self, final, encoderTensors):
        """Vanilla FPN Decoder

        INPUTS:
            - encoderTensors (list of SparseTensor): output of encoder.
        RETURNS:
            - decoderTensors (list of SparseTensor):
            list of feature tensors in decoding path at each spatial resolution.
        """
        decoderTensors = []
        x = final
        for i, layer in enumerate(self.decoding_conv):
            eTensor = encoderTensors[-i - 2]
            x = layer(x)
            x += self.lateral[i](eTensor)
            x = self.decoding_block[i](x)
            decoderTensors.append(x)
        return decoderTensors

    def forward(self, input):
        coords = input[:, 0 : self.D + 1].int()
        features = input[:, self.D + 1 :]

        x = ME.SparseTensor(features, coords=coords)
        encoderOutput = self.encoder(x)
        encoderTensors = encoderOutput["encoderTensors"]
        finalTensor = encoderOutput["finalTensor"]
        decoderTensors = self.decoder(finalTensor, encoderTensors)

        res = {"encoderTensors": encoderTensors, "decoderTensors": decoderTensors}
        return res



"""Analysis scripts and performance evaluation tools.

This module provides high-level analysis capabilities for neutrino physics
reconstruction results, building upon the output of the ML reconstruction
chain and post-processing modules.

**Core Analysis Management:**
- `AnaManager`: Central orchestrator for analysis workflows
  - Configuration-driven analysis pipeline
  - Multi-file dataset processing
  - Result aggregation and export
  - Performance metrics calculation

**Analysis Categories:**

**Reconstruction Performance:**
- Particle identification efficiency and purity
- Energy resolution and bias studies
- Spatial resolution metrics
- Angular reconstruction accuracy
- Vertex finding performance

**Physics Analysis:**
- Event topology classification
- Neutrino interaction type identification
- Cross-section measurements
- Background rejection studies
- Signal efficiency optimization

**Quality Assessment:**
- Truth matching quality metrics
- Reconstruction completeness
- Fragment and particle merging/splitting rates
- Detector systematic effects
- Data/MC comparison studies

**Output Products:**
- Statistical summaries and histograms
- Performance plots and visualizations
- Efficiency/purity curves
- Resolution studies
- Publication-ready figures
- Analysis reports and tables

**Workflow Integration:**
```
ML Chain → Post-Processing → Analysis → Physics Results
    ↓              ↓              ↓           ↓
 Raw Reco      High-level     Metrics     Publications
 Objects       Physics        Plots       Papers
              Quantities     Tables      Talks
```

**Key Features:**
- Automated truth matching and validation
- Configurable analysis cuts and selections
- Statistical uncertainty propagation
- Systematic error evaluation
- Batch processing capabilities
- Interactive analysis notebooks
- Custom metric definitions

**Example Usage:**
```python
from spine.ana import AnaManager

# Initialize analysis with configuration
ana = AnaManager(config_file='analysis.yaml')

# Run full analysis pipeline
results = ana.analyze(input_files, output_dir)

# Generate performance plots
ana.plot_performance(results)
```

This module serves as the final step in the SPINE reconstruction pipeline,
transforming raw ML outputs into physics-ready analysis results.
"""

from .manager import AnaManager



"""Complete graph constructor for GNNs."""

import numba as nb
import numpy as np

from .base import GraphBase

__all__ = ["CompleteGraph"]


class CompleteGraph(GraphBase):
    """Generates graphs that connect each node with every other node.

    If two nodes belong to separate batches, they cannot be connected.

    See :class:`GraphBase` for attributes/methods shared
    across all graph constructors.
    """

    # Name of the graph constructor (as specified in the configuration)
    name = "complete"

    def generate(self, clusts, **kwargs):
        """Generates a complete graph on a set of batched nodes.

        Parameters
        ----------
        clusts : IndexBatch
            (C) Cluster indexes
        **kwargs : dict, optional
            Unused graph generation arguments

        Returns
        -------
        np.ndarray
            (2, E) Tensor of edges
        np.ndarray
            (B) Number of edges in each entry of the batch
        """
        return self._generate(clusts.counts)

    @staticmethod
    @nb.njit(cache=True)
    def _generate(counts: nb.int64[:]) -> (nb.int64[:, :], nb.int64[:]):
        # Loop over the batches, define the adjacency matrix for each
        edge_counts = counts * (counts - 1) // 2
        num_edges = np.sum(edge_counts)
        edge_index = np.empty((2, num_edges), dtype=np.int64)
        offset, index = 0, 0
        for b in range(len(counts)):
            # Build a list of edges
            c = counts[b]
            adj_mat = np.triu(np.ones((c, c)), k=1)
            edges = np.vstack(np.where(adj_mat))
            num_edges_b = edges.shape[1]

            edge_index[:, index : index + num_edges_b] = offset + edges
            index += num_edges_b
            offset += c

        return edge_index, edge_counts



"""Custom non-linear activation functions."""

import MinkowskiEngine as ME
import torch


class MinkowskiMish(torch.nn.Module):
    """Mish non-linearity layer.

    Reference: https://arxiv.org/pdf/1908.08681.pdf
    """

    def __init__(self):
        """Initialize the layer."""
        super().__init__()

    def forward(self, input_data):
        """Pass tensor through the layer.

        Parameters
        ----------
        input_data : ME.SparseTensor
            Sparse input tensor

        Return
        ------
        ME.SparseTensor
            Sparse output tensor
        """
        out = torch.nn.functional.softplus(input_data.F)
        out = torch.tanh(out)
        out = out * input_data.F

        return ME.SparseTensor(
            out, coords_key=input_data.coords_key, coords_manager=input_data.coords_man
        )



"""Analysis script used to evaluate the point proposal accuracy."""

import numpy as np
from scipy.spatial.distance import cdist

from spine.ana.base import AnaBase
from spine.utils.globals import (
    COORD_COLS,
    LOWES_SHP,
    PPN_END_COLS,
    PPN_LENDP_COL,
    PPN_LTYPE_COL,
    PPN_SHAPE_COL,
)

__all__ = ["PointProposalAna"]


class PointProposalAna(AnaBase):
    """Class which computes and stores the necessary data to evaluate the
    point proposal accuracy.

    It evaluates the following metrics:
    - Distance from true to closest predicted point (efficiency)
    - Distance from predicted to closest true point (purity)
    - Point type classification accuracy
    - Point end classification accuracy
    """

    # Name of the analysis script (as specified in the configuration)
    name = "point_eval"

    # Set of data keys needed for this analysis script to operate
    _keys = (("ppn_pred", True),)

    def __init__(
        self, num_classes=LOWES_SHP, label_key="ppn_label", endpoints=False, **kwargs
    ):
        """Initialize the analysis script.

        Parameters
        ----------
        num_classes : int, default 4
            Number of pixel classses, excluding the ghost class
        label_key : str, default 'seg_label'
            Name of the tensor which contains the segmentation labels
        endpoints : bool, default False
            Evaluate the accuracy of end point classification
        **kwargs : dict, optional
            Additional arguments to pass to :class:`AnaBase`
        """
        # Initialize the parent class
        super().__init__(**kwargs)

        # Store the basic parameters
        self.num_classes = num_classes
        self.label_key = label_key
        self.endpoints = endpoints

        # Append other required key
        self.update_keys({self.label_key: True})

        # Initialize the output
        self.initialize_writer("truth_to_reco")
        self.initialize_writer("reco_to_truth")

        # Initialize a dummy dictionary to return when there is no match
        self.dummy_dict = {"dist": -1.0, "shape": -1, "closest_shape": -1}
        for s in range(self.num_classes):
            self.dummy_dict[f"dist_{s}"] = -1.0
        if endpoints:
            self.dummy_dict["end"] = -1
            self.dummy_dict["closest_end"] = -1

    def process(self, data):
        """Store the semantic segmentation metrics for one entry.

        Parameters
        ----------
        data : dict
            Dictionary of data products
        """
        # Initialize dictionaries to compare
        points, types, ends = {}, {}, {}

        # Fetch the list of label points and their characteristics
        points["truth"] = data[self.label_key][:, COORD_COLS]
        types["truth"] = data[self.label_key][:, PPN_LTYPE_COL].astype(int)
        if self.endpoints:
            ends["truth"] = data[self.label_key][:, PPN_LENDP_COL].astype(int)

        # Fetch the list of predicted points and their characteristics
        points["reco"] = data["ppn_pred"][:, COORD_COLS]
        types["reco"] = data["ppn_pred"][:, PPN_SHAPE_COL].astype(int)
        if self.endpoints:
            ends["reco"] = np.argmax(data["ppn_pred"][:, PPN_END_COLS], axis=1)

        # Compute the pair-wise distances between label and predicted points
        dist_mat = {}
        dist_mat["reco"] = cdist(points["reco"], points["truth"])
        dist_mat["truth"] = dist_mat["reco"].T

        # Store one row per predicted point
        matches = (("truth", "reco"), ("reco", "truth"))
        for source, target in matches:
            # If there is no source points, nothing to do
            if len(points[source]) == 0:
                continue

            # If there are no target points, record no match
            if len(points[target]) == 0:
                # Append dummy values
                dummy = {**self.dummy_dict}
                for i in range(len(points[source])):
                    dummy["shape"] = types[source][i]
                    if self.endpoints:
                        dummy["end"] = ends[source][i]
                    self.append(f"{source}_to_{target}", **dummy)

                # Proceed
                continue

            # Otherwise, use closest point as reference
            dists = dist_mat[source]
            closest_index = np.argmin(dists, axis=1)
            masks = [np.where(types[target] == s)[0] for s in range(self.num_classes)]
            for i in range(len(points[source])):
                point_dict = {**self.dummy_dict}
                point_dict["dist"] = dists[i, closest_index[i]]
                point_dict["shape"] = types[source][i]
                point_dict["closest_shape"] = types[target][closest_index[i]]
                if self.endpoints:
                    point_dict["end"] = ends[source][i]
                    point_dict["closest_end"] = ends[target][closest_index[i]]
                for s in range(self.num_classes):
                    if len(masks[s]) > 0:
                        point_dict[f"dist_{s}"] = np.min(dists[i, masks[s]])

                self.append(f"{source}_to_{target}", **point_dict)



"""Routines to produce reconstruction performance metric plots."""

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.ticker import Formatter

__all__ = ["heatmap", "annotate_heatmap"]


class UncertaintyFormatter(Formatter):
    """Use a new-style format string (as used by `str.format`) to format the tick.

    The field used for the tick value must be labeled *x* and the field used
    for the tick position must be labeled *pos*.
    """

    def __init__(self, fmt):
        self.fmt = fmt

    def __call__(self, x, pos=None):
        """
        Return the formatted label string.

        *x* and *pos* are passed to `str.format` as keyword arguments
        with those exact names.
        """
        return self.fmt.format(*x)


def heatmap(data, row_labels, col_labels, ax=None, **kwargs):
    """Create a heatmap from a numpy array and two lists of labels.

    Parameters
    ----------
    data
        A 2D numpy array of shape (N, M).
    row_labels
        A list or array of length N with the labels for the rows.
    col_labels
        A list or array of length M with the labels for the columns.
    ax
        A `matplotlib.axes.Axes` instance to which the heatmap is plotted.  If
        not provided, use current axes or create a new one.  Optional.
    **kwargs
        All other arguments are forwarded to `imshow`.
    """

    if not ax:
        ax = plt.gca()

    # Plot the heatmap
    im = ax.imshow(data, **kwargs)

    # We want to show all ticks...
    ax.set_xticks(np.arange(data.shape[1]))
    ax.set_yticks(np.arange(data.shape[0]))
    # ... and label them with the respective list entries.
    ax.set_xticklabels(col_labels, fontsize=18)
    ax.set_yticklabels(row_labels, fontsize=18)

    # Let the horizontal axes labeling appear on top.
    ax.tick_params(top=True, bottom=False, labeltop=True, labelbottom=False)

    # # Rotate the tick labels and set their alignment.
    plt.setp(ax.get_xticklabels(), rotation=0, ha="center", rotation_mode="anchor")

    # Turn spines off and create white grid.
    ax.spines[:].set_visible(False)

    ax.set_xticks(np.arange(data.shape[1] + 1) - 0.5, minor=True)
    ax.set_yticks(np.arange(data.shape[0] + 1) - 0.5, minor=True)
    ax.grid(which="minor", color="w", linestyle="-", linewidth=3)
    ax.tick_params(which="minor", bottom=False, left=False)

    return im


def annotate_heatmap(
    im,
    data=None,
    unc=None,
    valfmt="{x:.2f}",
    textcolors=("black", "white"),
    threshold=None,
    **textkw,
):
    """A function to annotate a heatmap.

    Parameters
    ----------
    im
        The AxesImage to be labeled.
    data
        Data used to annotate.  If None, the image's data is used.  Optional.
    valfmt
        The format of the annotations inside the heatmap.  This should either
        use the string format method, e.g. "$ {x:.2f}", or be a
        `matplotlib.ticker.Formatter`.  Optional.
    textcolors
        A pair of colors.  The first is used for values below a threshold,
        the second for those above.  Optional.
    threshold
        Value in data units according to which the colors from textcolors are
        applied.  If None (the default) uses the middle of the colormap as
        separation.  Optional.
    **kwargs
        All other arguments are forwarded to each call to `text` used to create
        the text labels.
    """

    if not isinstance(data, (list, np.ndarray)):
        data = im.get_array()

    # Normalize the threshold to the images color range.
    if threshold is not None:
        threshold = im.norm(threshold)
    else:
        threshold = im.norm(data.max()) / 2.0

    # Set default alignment to center, but allow it to be
    # overwritten by textkw.
    kw = {"horizontalalignment": "center", "verticalalignment": "center"}
    kw.update(textkw)

    # Get the formatter in case a string is supplied
    if isinstance(valfmt, str):
        valfmt = UncertaintyFormatter(valfmt)

    # Loop over the data and create a `Text` for each "pixel".
    # Change the text's color depending on the data.
    texts = []
    for i in range(data.shape[0]):
        for j in range(data.shape[1]):
            kw.update(color=textcolors[int(im.norm(data[i, j]) > threshold)])
            text = im.axes.text(j, i, valfmt((data[i, j], unc[i, j]), None), **kw)
            texts.append(text)

    return texts



"""Module that defines a generic node classification loss."""

from warnings import warn

import numpy as np
import torch

from spine.data import IndexBatch
from spine.model.layer.factories import loss_fn_factory
from spine.utils.globals import (
    COORD_END_COLS,
    COORD_START_COLS,
    PART_COL,
    SHAPE_COL,
    TRACK_SHP,
)
from spine.utils.gnn.cluster import get_cluster_label_batch

__all__ = ["NodeOrientLoss"]


class NodeOrientLoss(torch.nn.Module):
    """Loss to learn how to point a track node in the right direction.

    Takes the 2-channel node output of the GNN and optimizes node-wise scores
    such that the score corresponding to the correct orientation is maximized.

    For use in config:

    ..  code-block:: yaml

        model:
          name: grappa
          modules:
            grappa_loss:
              node_loss:
                name: orient
                <dictionary of arguments to pass to the loss>

    See configuration files prefixed with `grappa_` under the `config`
    directory for detailed examples of working configurations.
    """

    # Name of the loss (as specified in the configuration)
    name = "orient"

    # Alternative allowed names of the loss
    aliases = ("orientation",)

    def __init__(self, loss="ce"):
        """Initialize the node orientation loss function.

        Parameters
        ----------
        loss : str, default 'ce'
            Name of the loss function to apply
        """
        # Initialize the parent class
        super().__init__()

        # Set the loss
        self.loss_fn = loss_fn_factory(loss, functional=True)

    def forward(
        self,
        clust_label,
        coord_label,
        clusts,
        node_pred,
        start_points,
        end_points,
        **kwargs,
    ):
        """Applies the node orientation loss to a batch of data.

        Parameters
        ----------
        clust_label : TensorBatch
            (N, 1 + D + N_f) Tensor of cluster labels for the batch
        coord_label : TensorBatch, optional
            (P, 1 + D + 8) Tensor of start/end point labels for each
            true particle in the image
        clusts : IndexBatch
            (C) Index which maps each cluster to a list of voxel IDs
        node_pred : TensorBatch
            (C, 2) Node prediction logits (binary output)
        start_points : TensorBatch
            (C, 3) Start point features associated with each node
        end_points : TensorBatch
            (C, 3) End point features associated with each node
        **kwargs : dict, optional
            Other labels/outputs of the model which are not relevant here

        Returns
        -------
        loss : torch.Tensor
            Value of the loss
        accuracy : float
            Value of the node-wise orientation accuracy
        count : int
            Number of nodes the loss was applied to
        """
        # Fetch the true particle associations and the shape
        part_ids = get_cluster_label_batch(clust_label, clusts, column=PART_COL)
        global_part_ids = np.empty_like(part_ids.tensor, dtype=np.int64)
        for b in range(part_ids.batch_size):
            shift_part_ids = part_ids[b]
            valid_index = np.where(shift_part_ids > -1)[0]
            shift_part_ids[valid_index] += coord_label.edges[b].item()
            lower, upper = part_ids.edges[b], part_ids.edges[b + 1]
            global_part_ids[lower:upper] = shift_part_ids

        # Restrict the loss to matched track clusters
        shapes = get_cluster_label_batch(clust_label, clusts, column=SHAPE_COL)
        valid_index = np.where((global_part_ids > -1) & (shapes.tensor == TRACK_SHP))[0]

        # Fetch the true directions from the particle associations
        all_cols = np.concatenate((COORD_START_COLS, COORD_END_COLS))
        index = global_part_ids[valid_index]
        true_starts, true_ends = coord_label.tensor[index][:, all_cols].split(3, dim=1)
        true_dirs = true_ends - true_starts

        # Restrict the start/end points, compute the vector
        start_points = start_points.tensor[valid_index]
        end_points = end_points.tensor[valid_index]
        feat_dirs = end_points - start_points

        # For each node, check whether the vector that joins the start to end
        # point node features are aligned with the ground truth
        node_assn = torch.sign(torch.sum(true_dirs * feat_dirs, dim=1)).long()
        node_assn = (node_assn + 1) // 2

        # Compute the loss
        node_pred = node_pred.tensor[valid_index]
        loss = self.loss_fn(node_pred, node_assn, reduction="mean")

        # Compute accuracy of assignment (fraction of correctly assigned nodes)
        acc = 1.0
        if len(valid_index):
            acc = float(torch.sum(torch.argmax(node_pred, dim=1) == node_assn))
            acc /= len(valid_index)

        return {"accuracy": acc, "loss": loss, "count": len(valid_index)}



"""Classes in charge of constructing Particle objects."""

import numpy as np
from scipy.special import softmax

from spine.data.out import RecoParticle, TruthParticle
from spine.utils.globals import COORD_COLS, GROUP_COL, TRACK_SHP, VALUE_COL
from spine.utils.gnn.network import filter_invalid_nodes

from .base import BuilderBase

__all__ = ["ParticleBuilder"]


class ParticleBuilder(BuilderBase):
    """Builds reconstructed and truth particles.

    It takes the raw output of the reconstruction chain, extracts the
    necessary information and builds :class:`RecoParticle` and
    :class:`TruthParticle` objects from it.
    """

    # Builder name
    name = "particle"

    # Types of objects constructed by the builder
    _reco_type = RecoParticle
    _truth_type = TruthParticle

    # Necessary/optional data products to build a reconstructed object
    _build_reco_keys = (
        ("particle_clusts", True),
        ("particle_shapes", True),
        ("particle_start_points", True),
        ("particle_end_points", True),
        ("particle_group_pred", True),
        ("particle_node_type_pred", True),
        ("particle_node_primary_pred", True),
        ("particle_node_orient_pred", False),
        ("reco_fragments", False),
        *BuilderBase._build_reco_keys,
    )

    # Necessary/optional data products to build a truth object
    _build_truth_keys = (
        ("particles", False),
        ("graph_label", False),
        ("truth_fragments", False),
        *BuilderBase._build_truth_keys,
    )

    # Necessary/optional data products to load a reconstructed object
    _load_reco_keys = (
        ("reco_particles", True),
        ("reco_fragments", False),
        *BuilderBase._load_reco_keys,
    )

    # Necessary/optional data products to load a truth object
    _load_truth_keys = (
        ("truth_particles", True),
        ("truth_fragments", False),
        *BuilderBase._load_truth_keys,
    )

    def build_reco(self, data):
        """Builds :class:`RecoParticle` objects from the full chain output.

        Parameters
        ----------
        data : dict
            Dictionary of data products
        """
        return self._build_reco(**data)

    def _build_reco(
        self,
        points,
        depositions,
        particle_clusts,
        particle_shapes,
        particle_start_points,
        particle_end_points,
        particle_group_pred,
        particle_node_type_pred,
        particle_node_primary_pred,
        particle_node_orient_pred=None,
        sources=None,
        reco_fragments=None,
    ):
        """Builds :class:`RecoParticle` objects from the full chain output.

        Parameters
        ----------
        points : np.ndarray
            (N, 3) Set of deposition coordinates in the image
        depositions : np.ndarray
            (N) Set of deposition values
        particle_clusts : List[np.ndarray]
            (P) List of indexes, each corresponding to a particle instance
        particle_shapes : np.ndarray
            (P) List of particle shapes (shower, track, etc.)
        particle_start_points : np.ndarray
            (P, 3) List of particle start point coordinates
        particle_end_points : np.ndarray
            (P, 3) List of particle end point coordinates
        particle_group_pred : np.ndarray
            (P) Interaction group each particle belongs to
        particle_node_type_pred : np.ndarray
            (P, N_c) Particle identification logits
        particle_node_primary_pred : np.ndarray
            (P, 2) Particle primary classification logits
        particle_node_orient_pred : np.ndarray, optional
            (P, 2) Particle orientation classification logits
        sources : np.ndarray, optional
            (N, 2) Tensor which contains the module/tpc information
        reco_fragments : List[RecoFragment], optional
            (F) List of reconstructed fragments

        Returns
        -------
        List[RecoParticle]
            List of constructed reconstructed particle instances
        """
        # Convert the logits to softmax scores and the scores to a prediction
        pid_scores = softmax(particle_node_type_pred, axis=1)
        primary_scores = softmax(particle_node_primary_pred, axis=1)
        pid_pred = np.argmax(pid_scores, axis=1)
        primary_pred = np.argmax(primary_scores, axis=1)
        if particle_node_orient_pred is not None:
            orient_pred = np.argmax(particle_node_orient_pred, axis=1)

        # Prepare fragment associations, if they were built
        if reco_fragments is not None:
            particle_ids = [frag.particle_id for frag in reco_fragments]
            particle_ids = np.unique(particle_ids, return_inverse=True)[-1]

        # Loop over the particle instances
        reco_particles = []
        for i, index in enumerate(particle_clusts):
            # Initialize
            particle = RecoParticle(
                id=i,
                interaction_id=particle_group_pred[i],
                shape=particle_shapes[i],
                index=index,
                points=points[index],
                depositions=depositions[index],
                pid=pid_pred[i],
                primary_scores=primary_scores[i],
                is_primary=bool(primary_pred[i]),
            )

            # Set the PID scores without modifying the default size
            particle.pid_scores[: len(pid_scores[i])] = pid_scores[i]
            particle.pid_scores[len(pid_scores[i]) :] = 0.0

            # Set the end points
            particle.start_point = particle_start_points[i]
            if particle.shape == TRACK_SHP:
                particle.end_point = particle_end_points[i]

            # If the orientation prediction is provided, use it
            if orient_pred is not None and not orient_pred[i]:
                if particle.shape == TRACK_SHP:
                    particle.start_point, particle.end_point = (
                        particle.end_point,
                        particle.start_point,
                    )

            # Add optional arguments
            if sources is not None:
                particle.sources = sources[index]

            # Build fragment associations, if available
            if reco_fragments is not None:
                fragment_ids = np.where(particle_ids == i)[0]
                particle.fragments = [reco_fragments[j] for j in fragment_ids]
                particle.fragment_ids = fragment_ids
                for frag in particle.fragments:
                    frag.particle_id = i
                    frag.interaction_id = particle.interaction_id

            # Append
            reco_particles.append(particle)

        return reco_particles

    def build_truth(self, data):
        """Builds :class:`TruthParticle` objects from the full chain output.

        Parameters
        ----------
        data : dict
            Dictionary of data products
        """
        return self._build_truth(**data)

    def _build_truth(
        self,
        particles,
        label_tensor,
        points_label,
        depositions_label,
        depositions_q_label=None,
        label_adapt_tensor=None,
        points=None,
        depositions=None,
        label_g4_tensor=None,
        points_g4=None,
        depositions_g4=None,
        sources_label=None,
        sources=None,
        graph_label=None,
        truth_fragments=None,
    ):
        """Builds :class:`TruthParticle` objects from the full chain output.

        Parameters
        ----------
        particles : List[Particle]
            List of true particles
        label_tensor : np.ndarray
            Tensor which contains the cluster labels of each deposition
        points_label : np.ndarray
            (N', 3) Set of deposition coordinates in the label image (identical
            for pixel TPCs, different if deghosting is involved)
        depositions_label : np.ndarray
            (N') Set of true deposition values in MeV
        depositions_q_label : np.ndarray, optional
            (N') Set of true deposition values in ADC, if relevant
        label_adapt_tensor : np.ndarray, optional
            Tensor which contains the cluster labels of each deposition,
            adapted to the semantic segmentation prediction.
        points : np.ndarray, optional
            (N, 3) Set of deposition coordinates in the image
        depositions : np.ndarray, optional
            (N) Set of deposition values
        label_tensor_g4 : np.ndarray, optional
            Tensor which contains the cluster labels of each deposition
            in the Geant4 image (before the detector simulation)
        points_g4 : np.ndarray, optional
            (N'', 3) Set of deposition coordinates in the Geant4 image
        depositions_g4 : np.ndarray, optional
            (N'') Set of deposition values in the Geant4 image
        sources_label : np.ndarray, optional
            (N', 2) Tensor which contains the label module/tpc information
        sources : np.ndarray, optional
            (N, 2) Tensor which contains the module/tpc information
        graph_label : np.ndarray, optional
            (E, 2) Parentage relations in the set of particles
        truth_fragments : List[TruthFragment], optional
            (F) List of true fragments

        Returns
        -------
        List[TruthParticle]
            List of restored true particle instances
        """
        # Fetch the group ID of each of the particles
        group_ids = np.array([p.group_id for p in particles], dtype=int)

        # Prepare fragment associations, if they were built
        if truth_fragments is not None:
            fragment_group_ids = np.array(
                [frag.orig_group_id for frag in truth_fragments]
            )

        # Loop over the true *visible* particle instance groups
        truth_particles = []
        unique_group_ids = np.unique(label_tensor[:, GROUP_COL]).astype(int)
        valid_group_ids = unique_group_ids[unique_group_ids > -1]
        for i, group_id in enumerate(valid_group_ids):
            # Load the MC particle information
            assert group_id < len(
                particles
            ), "Invalid group ID, cannot build true particle."
            particle = TruthParticle(**particles[group_id].as_dict())
            assert (
                particle.id == group_id
            ), "The ordering of the true particles is wrong."

            # Override the index of the particle and its group, but preserve it
            particle.orig_id = group_id
            particle.orig_group_id = group_id
            particle.orig_parent_id = particle.parent_id
            particle.orig_interaction_id = particle.interaction_id
            particle.orig_children_id = particle.children_id

            particle.id = i
            particle.group_id = i
            particle.parent_id = i
            particle.children_id = np.empty(0, dtype=particle.orig_children_id.dtype)

            # Update the deposited energy attribute by summing that of all
            # particles in the group (LArCV definition != SPINE definition)
            particle.energy_deposit = 0.0
            for j in np.where(group_ids == group_id)[0]:
                particle.energy_deposit += particles[j].energy_deposit

            # Update the attributes shared between reconstructed and true
            particle.length = particle.distance_travel
            particle.is_primary = bool(particle.interaction_primary > 0)
            particle.start_point = particle.first_step
            if particle.shape == TRACK_SHP:
                particle.end_point = particle.last_step

            # Update the particle with its long-form attributes
            index = np.where(label_tensor[:, GROUP_COL] == group_id)[0]
            particle.index = index
            particle.points = points_label[index]
            particle.depositions = depositions_label[index]
            if depositions_q_label is not None:
                particle.depositions_q = depositions_q_label[index]
            if sources_label is not None:
                particle.sources = sources_label[index]

            if label_adapt_tensor is not None:
                index_adapt = np.where(label_adapt_tensor[:, GROUP_COL] == group_id)[0]
                particle.index_adapt = index_adapt
                particle.points_adapt = points[index_adapt]
                particle.depositions_adapt = depositions[index_adapt]
                if sources is not None:
                    particle.sources_adapt = sources[index_adapt]

            if label_g4_tensor is not None:
                index_g4 = np.where(label_g4_tensor[:, GROUP_COL] == group_id)[0]
                particle.index_g4 = index_g4
                particle.points_g4 = points_g4[index_g4]
                particle.depositions_g4 = depositions_g4[index_g4]

            # Build fragment associations, if available
            if truth_fragments is not None:
                fragment_ids = np.where(fragment_group_ids == group_id)[0]
                particle.fragments = [truth_fragments[j] for j in fragment_ids]
                particle.fragment_ids = fragment_ids
                for frag in particle.fragments:
                    frag.particle_id = i

            # Append
            truth_particles.append(particle)

        # If the parentage relations of non-empty particles are available,
        # use them to assign parent/children IDs in the new particle set
        if graph_label is not None:
            # Narrow down the list of edges to those connecting visible particles
            inval = set(np.unique(graph_label)).difference(set(valid_group_ids))
            if len(inval) > 0:
                graph_label = filter_invalid_nodes(graph_label, tuple(inval))

            # Use the remaining edges to build parentage relations
            mapping = {group_id: i for i, group_id in enumerate(valid_group_ids)}
            for source, target in graph_label:
                parent = truth_particles[mapping[source]]
                child = truth_particles[mapping[target]]

                child.parent_id = parent.id
                parent.children_id = np.append(parent.children_id, child.id)

        return truth_particles

    def load_reco(self, data):
        """Construct :class:`RecoParticle` objects from their stored versions.

        Parameters
        ----------
        data : dict
            Dictionary of data products
        """
        return self._load_reco(**data)

    def _load_reco(
        self,
        reco_particles,
        points=None,
        depositions=None,
        sources=None,
        reco_fragments=None,
    ):
        """Construct :class:`RecoParticle` objects from their stored versions.

        Parameters
        ----------
        reco_particles : List[RecoParticle]
            (P) List of partial reconstructed particles
        points : np.ndarray, optional
            (N, 3) Set of deposition coordinates in the image
        depositions : np.ndarray, optional
            (N) Set of deposition values
        sources : np.ndarray, optional
            (N, 2) Tensor which contains the module/tpc information
        reco_fragments : List[RecoFragment], optional
            (F) List of reconstructed fragments

        Returns
        -------
        List[RecoParticle]
            List of restored reconstructed particle instances
        """
        # Loop over the dictionaries
        for i, particle in enumerate(reco_particles):
            # Check that the particle ID checks out
            assert particle.id == i, "The ordering of the stored particles is wrong."

            # Update the particle with its long-form attributes
            if points is not None:
                particle.points = points[particle.index]
                particle.depositions = depositions[particle.index]
                if sources is not None:
                    particle.sources = sources[particle.index]

            # Load the fragment associations, if available
            if reco_fragments is not None:
                particle.fragments = [reco_fragments[j] for j in particle.fragment_ids]

        return reco_particles

    def load_truth(self, data):
        """Construct :class:`TruthParticle` objects from their stored versions.

        Parameters
        ----------
        data : dict
            Dictionary of data products
        """
        return self._load_truth(**data)

    def _load_truth(
        self,
        truth_particles,
        points_label=None,
        depositions_label=None,
        depositions_q_label=None,
        points=None,
        depositions=None,
        points_g4=None,
        depositions_g4=None,
        sources_label=None,
        sources=None,
        truth_fragments=None,
    ):
        """Construct :class:`TruthParticle` objects from their stored versions.

        Parameters
        ----------
        truth_particles : List[TruthParticle]
            (P) List of partial truth particles
        points_label : np.ndarray, optional
            (N', 3) Set of deposition coordinates in the label image (identical
            for pixel TPCs, different if deghosting is involved)
        depositions_label : np.ndarray, optional
            (N') Set of true deposition values in MeV
        depositions_q_label : np.ndarray, optional
            (N') Set of true deposition values in ADC, if relevant
        points : np.ndarray, optional
            (N, 3) Set of deposition coordinates in the image
        depositions : np.ndarray, optional
            (N) Set of deposition values
        points_g4 : np.ndarray, optional
            (N'', 3) Set of deposition coordinates in the Geant4 image
        depositions_g4 : np.ndarray, optional
            (N'') Set of deposition values in the Geant4 image
        sources_label : np.ndarray, optional
            (N', 2) Tensor which contains the label module/tpc information
        sources : np.ndarray, optional
            (N, 2) Tensor which contains the module/tpc information
        truth_fragments : List[TruthFragment], optional
            (F) List of true fragments

        Returns
        -------
        List[TruthParticle]
            List of restored true particle instances
        """
        # Loop over the dictionaries
        for i, particle in enumerate(truth_particles):
            # Check that the particle ID checks out
            assert particle.id == i, "The ordering of the stored particles is wrong."

            # Update the particle with its long-form attributes
            if points_label is not None:
                particle.points = points_label[particle.index]
                particle.depositions = depositions_label[particle.index]
                if depositions_q_label is not None:
                    particle.depositions_q = depositions_q_label[particle.index]
                if sources_label is not None:
                    particle.sources = sources_label[particle.index]

            if points is not None:
                particle.points_adapt = points[particle.index_adapt]
                particle.depositions_adapt = depositions[particle.index_adapt]
                if sources is not None:
                    particle.sources_adapt = sources[particle.index_adapt]

            if points_g4 is not None:
                particle.points_g4 = points_g4[particle.index_g4]
                particle.depositions_g4 = depositions_g4[particle.index_g4]

            # Load the fragment associations, if available
            if truth_fragments is not None:
                particle.fragments = [truth_fragments[j] for j in particle.fragment_ids]

        return truth_particles



"""SPINE configuration loading system.

This package provides a sophisticated configuration loading system with:
- Hierarchical file includes with cycle detection
- Metadata support for versioning and configuration
- Override semantics with dot-notation
- Collection operations (list/dict)
- Configurable strict modes

Main Entry Point
----------------
load_config : Load a SPINE configuration file

See loader module docstring for full configuration language specification.
"""

from .api import API_VERSION
from .download import download_from_url, get_cache_dir
from .errors import (
    ConfigCycleError,
    ConfigError,
    ConfigIncludeError,
    ConfigOperationError,
    ConfigPathError,
    ConfigTypeError,
    ConfigValidationError,
)
from .load import load_config, load_config_file

__version__ = API_VERSION

__all__ = [
    "load_config",
    "load_config_file",
    "download_from_url",
    "get_cache_dir",
    "ConfigError",
    "ConfigIncludeError",
    "ConfigCycleError",
    "ConfigPathError",
    "ConfigTypeError",
    "ConfigOperationError",
    "ConfigValidationError",
]
